"""
    PeakCurve.py

        temporary safer replacement of ElutionCurve for PeakMapper

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from molass_legacy.KekLib.SciPyCookbook import smooth
import molass_legacy.KekLib.DebugPlot as plt
from GeometryUtils import rotated_argminmax
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
from OptimalSmoothing import OptimalSmoothing
from CurveFeatures import CurveFeatures

CORRECT_OK_LIMIT = 0.02     # > 0.015 for 20211209
NUM_CORRECTION = 2
END_WIDTH = 5
SAFE_SIZE_RATIO = 0.6       # < 0.65 for 20191118_3

def shift_range_pairs(j, range_pairs):
    return [ [ [ j + r for r in range_ ] for range_ in range_pair ] for range_pair in range_pairs ]

def get_peak_curve_info(curve, debug=False):
    from LPM import get_corrected   # moved due to ImportError: ... (most likely due to a circular import)

    x = curve.x
    y = curve.y
    max_y = curve.max_y

    start = 0
    for k in range(NUM_CORRECTION):
        cr_y = get_corrected(y)
        cm_y = np.cumsum(cr_y)  # note that this correction is for peak region recognition only
        if debug:
            with plt.Dp():
                fig,ax = plt.subplots()
                ax.set_title("get_peak_curve_info", fontsize=16)
                ax.plot(x, y, label="data")
                ax.plot(x, cr_y, ":", label="corrected")
                axt = ax.twinx()
                axt.grid(False)
                axt.plot(x, cm_y, color="cyan", label="cumulated")
                ax.legend(fontsize=16)
                axt.legend(loc="center right", fontsize=16)
                fig.tight_layout()
                plt.show()
        i1, i2 = rotated_argminmax(-np.pi/8, cm_y, debug=debug)
        half_width = (i2 - i1)*2//3
        center = (i1 + i2)//2
        j1 = max(0, center - half_width)
        j2 = min(len(x)-1, center+ half_width)
        slice_ = slice(j1, j2)
        x_ = x[slice_]
        y_ = cr_y[slice_]
        ratio1 = abs(np.average(y_[0:END_WIDTH])/max_y)
        ratio2 = abs(np.average(y_[-END_WIDTH:])/max_y)
        if debug:
            print([k], "ratio1=", ratio1, "ratio2=", ratio2)
        if ratio1 < CORRECT_OK_LIMIT and ratio2 < CORRECT_OK_LIMIT:
            break
        else:
            if k < NUM_CORRECTION - 1:
                start += j1
                x = x_
                y = y_

    prs = [int(round(p)) for p in curve.get_peak_region_sigma()]
    slice_ = slice(min(prs[0], start+j1), max(prs[1], start+j2))

    size_ratio = (slice_.stop - slice_.start)/len(curve.x)      # use curve.x not x for x may have been changed
    # print("------------------ size_ratio=", size_ratio)
    assert size_ratio < SAFE_SIZE_RATIO

    if debug:
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            ax1.plot(curve.x, curve.y, label="orignal data")
            ax1.plot(x, cr_y, label="corrected data")
            ax1.plot(x_, y_, ":", label="sliced data")
            ax2.plot(x_, y_, label="sliced data")
            for ax in ax1, ax2:
                ax.legend()
            fig.tight_layout()
            plt.show()

    if debug:
        with plt.Dp():
            fig,ax = plt.subplots()
            ax.set_title("get_peak_curve_info", fontsize=16)
            ax.plot(curve.x, curve.y, label="data")
            ax.plot(x, cr_y, ":", label="corrected")
            ymin,ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax)
            for j in [slice_.start, slice_.stop]:
                ax.plot([j, j], [ymin,ymax], color="yellow")
            axt = ax.twinx()
            axt.grid(False)
            axt.plot(x, cm_y, color="cyan", label="cumulated")
            ax.legend(fontsize=16)
            axt.legend(loc="center right", fontsize=16)
            fig.tight_layout()
            plt.show()

    ecurve = ElutionCurve(curve.y[slice_])      # use no correction y
    ecurve.get_default_editor_ranges()

    if debug:
        from molass_legacy.Elution.CurveUtils import simple_plot
        fig, ax = plt.subplots()
        ax.set_title("get_peak_curve_info")
        simple_plot(ax, ecurve)
        plt.show()

    return ecurve, slice_

class PeakCurve(ElutionCurve):
    def __init__(self, curve, pkslice_info=None, debug=False):
        self.logger = logging.getLogger(__name__)

        if pkslice_info is None:
            pkcurve, slice_ = get_peak_curve_info(curve, debug=debug)
        else:
            pkcurve, slice_ = pkslice_info

        self.peak_slice = slice_
        self.x = x = curve.x
        self.y = y = curve.y
        self.possiblly_peakless = False
        self.emg_peaks = None
        self.paired_ranges = None

        if debug:
            from molass_legacy.Elution.CurveUtils import simple_plot
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("sliced curve")
                simple_plot(ax, pkcurve)
                fig.tight_layout()
                plt.show()

        j1 = slice_.start
        self.min_peak_width = pkcurve.min_peak_width
        self.peak_top_x = j1 + pkcurve.peak_top_x
        self.peak_info = [[j1 + j for j in rec] for rec in pkcurve.peak_info]
        self.boundaries = [j1 + b for b in pkcurve.boundaries]
        self.major_peak_info = [[j1 + j for j in rec] for rec in pkcurve.major_peak_info]
        self.editor_ranges = shift_range_pairs(j1, pkcurve.editor_ranges)

        self.max_x = np.argmax(y)
        self.max_y = y[self.max_x]
        self.min_y = np.percentile(y, 5)
        self.height = self.max_y - self.min_y

        self.smoothing = OptimalSmoothing(x, y, self.height, self.min_y)
        self.smoothing.compute_optimal_curves()
        self.y_for_spline = self.smoothing.spline_y
        self.feature_y = self.smoothing.feature_y
        self.spline = self.smoothing.spline
        self.sy = self.spline(x)

        self.d1 = self.smoothing.d1
        self.d1y = self.d1(x)
        self.d2 = self.smoothing.d2
        self.d2y = self.d2(x)
        self.d2_roots = self.smoothing.d2_roots

        self.peak_region_sigma = None

        try:
            self.feature = CurveFeatures(self)
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            print(etb)

    def get_major_peak_info(self, ratio=None):
        return self.major_peak_info

    def get_peak_region_width(self):
        # note that this overrides ElutionCurve.get_peak_region_width()
        # or rather, ElutionCurve.get_peak_region_width() is a substitute for this
        # when a PeakCurve is not used
        return self.peak_slice.stop - self.peak_slice.start



# Add the root directory to the system path
# This is necessary to import the package correctly in the Sphinx documentation.
import sys
import os
root_dir = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
sys.path.insert(0, root_dir)

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'Molass Library Legacy'
copyright = '2025, The Molass Community'
author = 'Molass Community'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosectionlabel",
    'sphinx_copybutton',
    'myst_parser',
]

autoclass_content = 'both'
# napoleon_google_docstring = True
# napoleon_numpy_docstring = True

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# Avoid using "package" in titles
add_module_names = False  # Removes the "molass." prefix from module names in titles

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme_options = {
    "repository_url": "https://github.com/freesemt/molass-legacy",
    "use_repository_button": True,
}

html_theme = 'sphinx_book_theme'
html_static_path = ['_static']
html_logo = "_static/molamola.png"
html_favicon = "_static/molamola.png"


"""
    Test.TesterLogger.py

    Copyright (c) 2017-2022, SAXS Team, KEK-PF
"""

import os
import re
import queue
import threading
from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry, open_w_safely

tester_log_queue = None

log_fh = None
dev_fh = None

def open_tester_log( home_dir ):
    global log_fh

    log_dir     = os.path.abspath( home_dir + '/log' ).replace( '\\', '/' )
    if not os.path.exists( log_dir ):
        mkdirs_with_retry(log_dir  )

    log_path = os.path.join(log_dir, "tester.log")
    log_fh = open_w_safely( log_path, "w", rename_cb=rename_log )
    create_log_queue()

    return log_fh

def rename_log( path ):
    n = 0
    path_bak = path
    while os.path.exists( path_bak ):
        path_bak = re.sub( r"(-\d*)?\.log$", "-%02d.log" % n, path )
        n += 1
    os.rename( path, path_bak ) 

def write_to_log( recstr ):
    global log_fh

    lock = threading.Lock()
    with lock:
        for fh in [ log_fh, dev_fh ]:
            if fh is None:
                continue
            fh.write( recstr )
            fh.flush()

def create_log_queue():
    global tester_log_queue
    if tester_log_queue is None:
        tester_log_queue = queue.Queue()

def put_to_tester_log_queue( recstr ):
    global tester_log_queue
    if tester_log_queue is None: return

    tester_log_queue.put( recstr )

def write_to_tester_log( recstr ):
    put_to_tester_log_queue( recstr )

def write_from_log_queue():
    global tester_log_queue

    if tester_log_queue is None:
        return

    lock = threading.Lock()
    with lock:
        while not tester_log_queue.empty():
            write_to_log( tester_log_queue.get() )

def open_dev_log( path ):
    global dev_fh
    dev_fh  = open( path, "w" )

def write_to_dev_log( recstr  ):
    global dev_fh
    if dev_fh is None: return

    dev_fh.write( recstr )
    dev_fh.flush()

def close_dev_log():
    global dev_fh
    if dev_fh is None: return
    dev_fh.close()
    dev_fh = None



# coding: utf-8
"""
    MicrofluidicElution.py

    Copyright (c) 2018-2019, SAXS Team, KEK-PF
"""
import os
import glob
import re
import numpy as np
from FlowSimulator import FlowSimulator

class MicrofluidicElution(FlowSimulator):
    def __init__(self, filepath, num_files):
        FlowSimulator.__init__(self, filepath, num_files=num_files)
        self.set_params()

    def set_flow_changes_as_a_temp_fix(self, absorbance):
        fc = self.get_linear_slope_ends()
        absorbance.jump_j = fc[0]
        absorbance.jump_j_safe = fc[0]
        absorbance.right_jump_j = fc[1]
        absorbance.right_jump_j_safe = fc[1]

    def set_baseplane_params_as_a_temp_fix(self, absorbance):
        absorbance.baseplane_params = ( 0, 0, 0 )

    def create_flow_change_proxy(self):
        return SimpleFlowChangeProxy(self)

    def propose_background_range(self, size, xy=None, width=10, debug=False, auto=True):
        x, y = self.get_elution_data(size)
        ii = self.get_linear_slope_ends()
        slope_start = ii[0]
        if xy is not None:
            x, y = xy

        min_y = np.min(y)
        lim_y = min_y*0.8 + y[slope_start]*0.2
        bx = np.where(np.logical_and(y < lim_y, x < x[slope_start]))[0]
        length = len(bx)
        stop = bx[-1] - int(length*0.1)
        start = stop - width

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            import time
            print('length=', length)
            fig = plt.figure()
            ax = fig.gca()
            ax.plot(y)
            for i in ii:
                ax.plot(i, y[i], 'o', color='red')
            ax.plot(bx, y[bx], 'o', color='yellow')
            ax.plot([start, stop], y[[start, stop]], 'o', color='green')
            fig.tight_layout()
            if auto:
                plt.show(block=False)
                time.sleep(0.5)
            else:
                plt.show()

        return (start, stop)

def get_mtd_elution(in_folder, num_files):
    from SerialDataUtils import get_mtd_filename
    filename = get_mtd_filename(in_folder)

    if filename is None:
        mtd_elution = None
    else:
        try:
            from MicrofluidicElution import MicrofluidicElution
            mtd_elution = MicrofluidicElution(filename, num_files)
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            print(etb)
            mtd_elution = None
    assert mtd_elution is not None    

    return mtd_elution

class CurveSimilarityProxy:
    def __init__(self, me):
        self.size  = len(me.x)
        self.slope  = 1
        self.intercept = 0
        self.mapped_info = [np.array([1, 0]), [1, 1], None]

    def inverse_int_value(self, j):
        i = int( (j - self.intercept)/self.slope + 0.5 )
        return max(0, min(self.size-1, i))

class SimpleFlowChangeProxy:
    def __init__(self, me):
        self.size = len(me.x)
        fc = me.get_linear_slope_ends()
        fc_ = fc[0:2]
        self.fc_list = fc_
        self.fc_safe_list = fc_
        self.similarity = CurveSimilarityProxy(me)

    def get_flow_changes(self):
        return self.fc_safe_list

    def get_real_flow_changes(self):
        flow_changes_src = self.fc_safe_list
        flow_changes = []
        for fc in flow_changes_src:
            if fc == 0 or fc == self.size - 1:
                fc_ = None
            else:
                fc_ = fc
            flow_changes.append( fc_ )
        return flow_changes

    def get_raw_flow_changes(self):
        return self.fc_list



"""
    Optimizer.BasinHopping.py

    overriding the original to investigate

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import basinhopping
from scipy.optimize._basinhopping import RandomDisplacement, AdaptiveStepsize
from scipy._lib._util import check_random_state

class CustomTakestep(AdaptiveStepsize):
    def __init__(self, *args, **kwargs):
        self.history = []
        AdaptiveStepsize.__init__(self, *args, **kwargs)

    def take_step(self, x):
        self.nstep += 1
        self.nstep_tot += 1
        if self.nstep % self.interval == 0:
            self._adjust_step_size()
        ret_x = self.takestep(x)
        self.history.append(ret_x)
        return ret_x

class BasinHopping:
    def __init__(self):
        pass

    def minimize(self, func, x0, **kwargs):

        custom = kwargs.pop('custom', False)
        if custom:
            seed = kwargs.pop('seed', None)

            # set up the np.random generator
            rng = check_random_state(seed)

            take_step = kwargs.pop('take_step', None)
            if take_step is None:

                stepsize = kwargs.pop('stepsize', 0.5)
                interval = kwargs.pop('interval', 50)
                target_accept_rate = kwargs.pop('target_accept_rate', 0.5)
                stepwise_factor = kwargs.pop('stepwise_factor', 0.9)
                disp = kwargs.pop('disp', False)

                displace = RandomDisplacement(stepsize=stepsize, random_gen=rng)
                take_step_wrapped = CustomTakestep(displace, interval=interval,
                                                     accept_rate=target_accept_rate,
                                                     factor=stepwise_factor,
                                                     verbose=disp)
                take_step = take_step_wrapped
                self.take_step = take_step
        else:
            take_step = None

        return basinhopping(func, x0, take_step=take_step, **kwargs)

    def show_history(self, optimizer, load=False):
        if load:
            history_array = np.loadtxt('history_array.dat')
        else:
            assert hasattr(self.take_step, 'history')
            history_array = np.array(self.take_step.history)
            np.savetxt('history_array.dat', history_array)

        history_dict = {}
        for p in history_array:
            sp = optimizer.split_params_simple(p)
            for k, x in enumerate(sp):
                seq = history_dict.get(k)
                if seq is None:
                    history_dict[k] = seq = []
                seq.append(x)

        keys = history_dict.keys()
        num_seqs = len(keys)

        import molass_legacy.KekLib.DebugPlot as plt
        from matplotlib.gridspec import GridSpec

        param_names = ["xr_components", "xr_baseline", "rg", "mapping", "uv_components", "uv_baseline", "(c,d)", "sec_params", "(Ti, Np)"]
        heights_    = [       8,             1,          2,      1,           2,               2,          1,         2,           1     ]
        heights = heights_[0:num_seqs]
        total_heights = np.sum(heights)
        print("total_heights=", total_heights)

        with plt.Dp(scrollable=True):
            fig = plt.figure(figsize=(12, total_heights))
            fig.suptitle("Variations of Normalized Parameters", fontsize=20)
            gs = GridSpec(total_heights, 8)
            start = 0
            for k, h in enumerate(heights):
                x = np.array(history_dict[k])
                print("x.shape=", x.shape)
                if k == 0:
                    for j, name in enumerate(["h", "mu(tR)", "sigma", "tau"]):
                        s = slice(start, start+2)
                        ax0 = fig.add_subplot(gs[s,0])
                        ax1 = fig.add_subplot(gs[s,1:])
                        start += 2
                        ax0.set_axis_off()
                        ax0.text(0.5, 0.5, name, ha="center", va="center")
                        for y in x[:,:,j].T:
                            ax1.plot(y)
                        ymin, ymax = ax1.get_ylim()
                        ax1.set_ylim(min(0, ymin), max(10, ymax))
                else:
                    s = slice(start, start+h)
                    ax0 = fig.add_subplot(gs[s,0])
                    ax1 = fig.add_subplot(gs[s,1:])
                    start += h
                    ax0.set_axis_off()
                    ax0.text(0.5, 0.5, param_names[k], ha="center", va="center")
                    for y in x.T:
                        ax1.plot(y)
                    ymin, ymax = ax1.get_ylim()
                    ax1.set_ylim(min(0, ymin), max(10, ymax))

            fig.tight_layout()
            dp = plt.get_dp()
            def resize():
                from molass_legacy.KekLib.TkUtils import split_geometry
                dp.update()
                geometry = dp.geometry()
                w, h, x, y = split_geometry(geometry)
                print(geometry)
                dp.geometry("%dx%d+%d+%d" % (w, min(900, h), x, y))
            dp.after(100, resize)
            plt.show()



"""
    SecTheory.InitialGuess.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from scipy.interpolate import UnivariateSpline
from molass_legacy.KekLib.GeometryUtils import rotated_argminmax
import molass_legacy.KekLib.DebugPlot as plt
from .CumsumInverse import cumsum_inverse

class InitialGuess:
    def __init__(self, ecurve, rg_info, sec_info, localopt, lrf_src=None, debug=False):
        self.sec_info = sec_info

        if lrf_src is not None:
            secparams = lrf_src.guess_secparams()

        x = ecurve.x
        y = ecurve.y
        sy = ecurve.sy
        cum_y = np.cumsum(sy)
        cum_y /= cum_y[-1]
        cum_spline = UnivariateSpline(x, cum_y, s=0, ext=3)

        k = 12
        def sigmoid(z):
            return 1/(1 + np.exp(-k*(z - 0.5)))

        nc = localopt.nc
        d_init = sec_info.d_init
        t0_init = sec_info.t0_init
        self.zeros_w = np.zeros(nc)

        init_y = np.linspace(0, 1, nc+2)[1:-1]
        init_ys = sigmoid(init_y)
        print("init_ys=", init_ys)

        m1, m2 = rotated_argminmax(-np.pi/64, cum_y, debug=False)
        init_tr = cumsum_inverse(cum_spline, init_ys, *x[[m1, m2]])
        rho_ = min(1, rg_info.rgs[-1]/d_init)
        K = (init_tr[-1] - t0_init)/(1 - rho_*0.5)
        print("init_tr=", init_tr)
        self.init_tr =init_tr

        if debug:
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("Initial Guess", fontsize=20)
                axt = ax2.twinx()
                axt.grid(False)

                ax1.set_title("Proportion Change Sigmoid (k=%d)" % k, fontsize=16)
                ax2.set_title("Integral Curve Mapped Points", fontsize=16)

                z = np.linspace(0, 1, 20)
                ax1.plot(z, sigmoid(z))
                ax1.plot(init_y, init_ys, "o", color="orange")

                ax2.plot(x, y)
                ax2.plot(x, ecurve.spline(x), ":")
                ymin, ymax = ax2.get_ylim()
                ax2.set_ylim(ymin, ymax)

                for px in [t0_init, K]:
                    ax2.plot([px, px], [ymin, ymax], color="green")

                for tr in init_tr:
                    ax2.plot([tr, tr], [ymin, ymax], color="yellow")

                axt.plot(x, cum_y, color="cyan")
                axt.plot(init_tr, cum_spline(init_tr), "o", color="pink")
                # axt.plot(init_tr, init_ys, "o", color="red")
                # ax2.plot(init_tr, ret.x, "o", color="orange")

                ymin, ymax = axt.get_ylim()
                ax1.set_ylim(ymin, ymax)
                axt.set_ylim(ymin, ymax)
                xmin, xmax = ax1.get_xlim()
                ax1.set_xlim(xmin, xmax)

                for y_, ys_ in zip(init_y, init_ys):
                    ax1.plot([y_, y_], [ymin, ys_], ":", color="pink")
                    ax1.plot([y_, xmax], [ys_, ys_], ":", color="pink")

                xmin, xmax = axt.get_xlim()
                axt.set_xlim(xmin, xmax)
                for ys_, tr_ in zip(init_ys, init_tr):
                    axt.plot([xmin, tr_], [ys_, ys_], ":", color="pink")

                fig.tight_layout()
                ret = plt.show()

            if not ret:
                raise RuntimeError("InitialGuess: Debug plot was canceled.")

        self.init_rg = d_init * (1 - np.power((init_tr - t0_init)/K, 1/2) )
        init_w_ = np.max([self.zeros_w, ecurve.spline(init_tr)], axis=0)
        self.init_seccol_params = sec_info.get_sec_params(K)

        def objective_scales(p, get_curves=False, debug=False):
            all_params = np.concatenate([self.init_seccol_params, self.init_rg, p])
            return localopt.objective_all(all_params, get_curves=get_curves, debug=debug, debug_title="scales")

        # adjust the scale parameters so that the areas take the same size.
        # this adjustment makes the choice of elution models safe or irrelevant.
        # i.e., both pdf and scaled pdf can be used interchangeably
        cy_list, ty = objective_scales(init_w_, get_curves=True)
        area_ratio = np.sum(y)/np.sum(ty)
        init_w_ *= area_ratio

        if debug:
            print("---------------- init_seccol_params=", self.init_seccol_params)
            print("---------------- init_rg=", self.init_rg)
            print("---------------- init_w_=", init_w_)
            init_fv = objective_scales(init_w_, debug=debug)
            print("---------------- init_fv=", init_fv)
            if localopt.abort:
                raise RuntimeError("InitialGuess: Debug plot was canceled before optimization.")

        # method="Nelder-Mead"
        self.ret = minimize(objective_scales, init_w_)

        if debug:
            ret = self.ret
            print("---------------- weights optimization: ret.success=", ret.success)
            print("---------------- weights optimization: ret.message=", ret.message)

            objective_scales(ret.x, debug=debug)
            if localopt.abort:
                raise RuntimeError("InitialGuess: Debug plot was canceled after optimization.")

    def get_params(self):
        init_w = np.max([self.zeros_w, self.ret.x], axis=0)
        return np.concatenate([self.init_seccol_params, self.init_rg, init_w])


class InitialGuessAdapter:
    def __init__(self, ret):
        pass


"""
    Selective.StochasticAdapter.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Batch.DataBridgeUtils import get_databridge
from molass_legacy.Decomposer.FitRecord import FitRecord
from molass_legacy.Decomposer.ModelEvaluator import ModelEvaluator
from molass_legacy.Models.Stochastic.MonoporeUvScaler import adjust_to_uv_scales
from molass_legacy.Selective.PeakProxy import PeakProxy

def convert_to_stochastic_decomposition(decomp_editor, debug=False):
    from importlib import reload
    import molass_legacy.Models.Stochastic.Monopore
    reload(molass_legacy.Models.Stochastic.Monopore)
    from molass_legacy.Models.Stochastic.Monopore import Monopore
    import molass_legacy.Decomposer.UnifiedDecompResult
    reload(molass_legacy.Decomposer.UnifiedDecompResult)
    from molass_legacy.Decomposer.UnifiedDecompResult import UnifiedDecompResult

    print("convert_to_stochastic_decomposition")
    
    bridge = get_databridge(decomp_editor)
    lrf_src = bridge.get_lrf_source()
    xr_x = lrf_src.xr_x
    xr_y = lrf_src.xr_y
    uv_x = lrf_src.uv_x
    uv_y = lrf_src.uv_y
    x_curve = decomp_editor.mapper.x_curve  # note that this is baseline-corrected
    a_curve = decomp_editor.mapper.a_curve  # same as above
    model = Monopore(delayed=True)          # get this from the editor
    model.set_fx_for_height_ratio(xr_x)

    mnp_params, corrected_rgs = lrf_src.guess_monopore_params(debug=debug)

    # convert mnp_params to decomp_result
    # task: unify the bridging procedure below with that in Estimator.StcEstimator.py

    a, b = bridge.peak_params_set[-2:]
    uv_x_ = a*xr_x + b                          # measured x
    uv_y_ = a_curve.spline(a*x_curve.x + b)     # note that uv_y_ is consistent with uv_x and uv_y

    if debug:
        from molass_legacy.Elution.CurveUtils import simple_plot
        print("len(xr_x)=", len(xr_x), "len(uv_x)=", len(uv_x))
        print("len(x_curve.x)=", len(x_curve.x), "len(a_curve.x)=", len(a_curve.x))
        with plt.Dp():
            fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(12, 10))
            fig.suptitle("convert_to_stochastic_decomposition inputs")
            simple_plot(axes[0,0], a_curve)
            simple_plot(axes[0,1], x_curve)
            axes[1,0].plot(uv_x, uv_y)
            axes[1,0].plot(uv_x_, uv_y_, ":")
            axes[1,1].plot(xr_x, xr_y)
            fig.tight_layout()
            ret = plt.show()
        if not ret:
            return

    xr_scales = mnp_params[6:]
    uv_scales, uv_ty = adjust_to_uv_scales(xr_x, xr_y, uv_x_, uv_y_, mnp_params, corrected_rgs, debug=debug)

    opt_recs = []
    opt_recs_uv = []

    # xr_area = np.sum(y)
    # uv_area = np.sum(uv_y)

    for kno, (xr_scale, uv_scale, rg) in enumerate(zip(xr_scales, uv_scales, corrected_rgs)):
        xr_params = np.concatenate([[xr_scale], mnp_params[:6], [rg]])
        cy = model.func(xr_x, *xr_params)
        j = np.argmax(cy)
        peak = PeakProxy(top_x=xr_x[j], top_y=cy[j])
        opt_rec = FitRecord(kno, ModelEvaluator(model, xr_params, sign=1, accepts_real_x=True), 0, peak)
        opt_recs.append(opt_rec)
        uv_params = np.concatenate([[uv_scale], mnp_params[:6], [rg]])
        cy = model.func(uv_x_, *uv_params)
        j = np.argmax(cy)
        peak = PeakProxy(top_x=uv_x_[j], top_y=cy[j])       # this is doubtfull. currently not used?
        opt_rec_uv = FitRecord(kno, ModelEvaluator(model, uv_params, sign=1, accepts_real_x=True), 0, peak)
        opt_recs_uv.append(opt_rec_uv)

    decomp_result = UnifiedDecompResult(
        xray_to_uv=None,
        x_curve=x_curve, x=xr_x, y=xr_y,
        opt_recs=opt_recs,
        max_y_xray=x_curve.max_y,
        model_name=model.get_name(),
        uv_y=uv_y_,
        opt_recs_uv=opt_recs_uv,
        max_y_uv=a_curve.max_y,
        )
    
    if debug:
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("convert_to_stochastic_decomposition outputs")
            ax1.set_title("UV")
            ax1.plot(uv_x, uv_y, label="data")
            cy_list = []
            for rec in opt_recs_uv:
                cy = rec.evaluator(xr_x)
                cy_list.append(cy)
                ax1.plot(uv_x_, cy, ":", label="component-%d" % rec.kno)
            ty = np.sum(cy_list, axis=0)
            ax1.plot(uv_x_, ty, ":", color="red", lw=2, label="model total")
            ax1.legend()

            ax2.set_title("XR")
            ax2.plot(xr_x, xr_y, label="data")
            cy_list = []
            for rec in opt_recs:
                cy = rec.evaluator(xr_x)
                cy_list.append(cy)
                ax2.plot(xr_x, cy, ":", label="component-%d" % rec.kno)
            ty = np.sum(cy_list, axis=0)
            ax2.plot(xr_x, ty, ":", color="red", lw=2, label="model total")
            ax2.legend()

            fig.tight_layout()
            plt.show()

    decomp_result.set_area_proportions()
    decomp_result.remove_unwanted_elements()    # or decomp_result.set_proportions() if you should keep minor components as well

    if debug:
        decomp_result.get_range_edit_info(debug=debug)

    return decomp_result


# coding: utf-8
"""
    ScatteringCurvePlotter.py

    Copyright (c) 2019-2021, SAXS Team, KEK-PF
"""
import os
from bisect import bisect_right
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.widgets import Button
from matplotlib.patches import Rectangle
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib import get_color, NavigationToolbar
from SerialDataUtils import serial_np_loadtxt

DEFAULT_CENTER  = 0.15
DEFAULT_WIDTH   = 0.02

title_texts = ["Intensity", "Error", "Error / Intensity"]

class ScatteringCurvePlotter(Dialog):
    def __init__(self, parent):
        self.log_scale = 1
        self.show_flag = 0
        self.num_files = 0
        self.lay_over_scaled = False
        self.drawn_data_list = []
        Dialog.__init__(self, parent, "Scattering Curve Plotter", visible=False)

    def body(self, body_frame):
        tk_set_icon_portable(self)

        cframe = Tk.Frame(body_frame)
        cframe.pack()
        tbframe = Tk.Frame(body_frame)
        tbframe.pack(anchor=Tk.W)
        bframe = Tk.Frame(body_frame)
        bframe.pack(anchor=Tk.E)


        self.toggle_scale_btn = Tk.Button(bframe, command=self.toggle_scale)
        self.toggle_scale_btn.grid(row=0, column=0, padx=10)
        self.toggle_scale_btn.grid_forget()

        self.scale_btn_frame = Tk.Frame(bframe)
        self.scale_btn_frame.grid(row=0, column=1, padx=10)
        self.scale_btn_frame.grid_forget()
        scale_btn = Tk.Button(self.scale_btn_frame, text="Lay Over Scale", command=self.lay_over_scale)
        scale_btn.grid(row=0, column=0)
        scale_center_label = Tk.Label(self.scale_btn_frame, text=" adjusting at center")
        scale_center_label.grid(row=0, column=1, padx=5)
        self.scale_center = Tk.DoubleVar()
        self.scale_center.set(DEFAULT_CENTER)
        scale_center_entry = Tk.Entry(self.scale_btn_frame, textvariable=self.scale_center, width=6, justify=Tk.CENTER)
        scale_center_entry.grid(row=0, column=2)
        scale_width_label = Tk.Label(self.scale_btn_frame, text=" with width")
        scale_width_label.grid(row=0, column=3, padx=5)
        self.scale_width = Tk.DoubleVar()
        self.scale_width.set(DEFAULT_WIDTH)
        scale_width_entry = Tk.Entry(self.scale_btn_frame, textvariable=self.scale_width, width=6, justify=Tk.CENTER)
        scale_width_entry.grid(row=0, column=4)
        self.scale_center.trace('w', self.params_tracer)
        self.scale_width.trace('w', self.params_tracer)

        self.clear_btn = Tk.Button(bframe, text="Clear", command=self.clear)
        self.clear_btn.grid(row=0, column=2, padx=10)
        self.clear_btn.grid_forget()

        self.fig, self.axes = plt.subplots( nrows=1, ncols=3, figsize=(21, 7) )
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )
        self.clear()

        self.fig.tight_layout()
        # self.toolbar = NavigationToolbar( self.mpl_canvas, tbframe, show_mode=False )
        self.toolbar = NavigationToolbar( self.mpl_canvas, tbframe )
        self.toolbar.update()

        self.add_dnd_bind()

    def show_guide_text(self):

        self.guide_texts = []
        for ax in self.axes:
            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
            tx = (xmin + xmax)/2
            ty = (ymin + ymax)/2
            t = ax.text(tx, ty, "Drag and drop files here to plot", alpha=0.3, fontsize=30, ha='center')
            self.guide_texts.append(t)

    def params_tracer(self, *args):
        self.lay_over_scaled = False

    def add_dnd_bind(self):
        self.mpl_canvas_widget.register_drop_target("*")

        def dnd_handler(event):
            self.on_entry(event)

        self.mpl_canvas_widget.bind("<<Drop>>", dnd_handler)

    def show(self):
        self._show()

    def on_entry(self, event):
        files = event.data.split(' ')
        # print('on_entry:', files)

        self.draw(files)

    def draw(self, files=[]):
        if self.guide_texts is not None:
            for t in self.guide_texts:
                t.remove()
            self.guide_texts = None

        self.update_scale()
        scale_center = self.scale_center.get()
        scale_width = self.scale_width.get()

        for file in files:
            self.lay_over_scaled = False
            data, _ = serial_np_loadtxt(file)
            self.drawn_data_list.append([file, data])
            x = data[:,0]
            for k, ax in enumerate(self.axes):
                if k < 2:
                    y = data[:,k+1]
                else:
                    y = data[:,2]/data[:,1]
                ax.plot(x, y, label=file)
            self.num_files += 1
            if self.num_files == 2:
                possible_center = x[-1] - scale_width/2
                if possible_center < scale_center:
                    self.scale_center.set(possible_center)
                self.scale_btn_frame.grid(row=0, column=1, padx=10)

        for ax in self.axes:
            ax.legend()
        self.mpl_canvas.draw()

    def update_scale(self):
        if self.log_scale:
            scale = 'log'
            btn_text = 'To Linear'
        else:
            scale = 'linear'
            btn_text = 'To Log'

        for ax in self.axes:
            ax.set_yscale(scale)
        self.toggle_scale_btn.config(text=btn_text)

        self.toggle_scale_btn.grid(row=0, column=0, padx=10)
        self.clear_btn.grid(row=0, column=2, padx=10)

    def toggle_scale(self):
        self.log_scale = 1 - self.log_scale
        self.update_scale()
        self.mpl_canvas.draw()

    def lay_over_scale(self):
        if not self.lay_over_scaled and len(self.drawn_data_list) > 0:
            self.lay_over_scale_drawn_data()

        ax = self.axes[0]
        ax.cla()

        self.update_scale()

        for file, data in self.drawn_data_list:
            x = data[:,0]
            y = data[:,1]
            ax.plot(x, y, label=file)

        if self.lay_over_scaled:
            f, t = self.x[self.index[[0, -1]]]
            ymin, ymax = ax.get_ylim()
            p = Rectangle(
                    (f, ymin),  # (x,y)
                    t - f,   # width
                    ymax - ymin,    # height
                    facecolor   = 'red',
                    alpha       = 0.1,
                )
            ax.add_patch(p)

        ax.legend()
        self.mpl_canvas.draw()

    def lay_over_scale_drawn_data(self):
        _, data = self.drawn_data_list[0]
        x = data[:,0]
        q = self.scale_center.get()
        w = self.scale_width.get()
        i = np.where(np.logical_and(x > q-w/2, x < q+w/2))[0]
        y = data[:,1]
        target = np.average(y[i])

        for _, data in self.drawn_data_list[1:]:
            y = data[:,1]
            y *= target/np.average(y[i])

        self.lay_over_scaled = True
        self.x = x
        self.index = i

    def clear(self):
        for k, ax in enumerate(self.axes):
            ax.cla()
            ax.set_title(title_texts[k], fontsize=20)
        self.drawn_data_list = []
        self.show_guide_text()
        self.mpl_canvas.draw()



# coding: utf-8
"""
    ConvolutionByDice.py

    Copyright (c) 2020,2024, SAXS Team, KEK-PF
"""
import numpy as np
# import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.patches import Polygon
import matplotlib.animation as animation
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.OurMatplotlib import get_color
from .DiscretePdf import DiscretePdf

def create_vertices(i, j):
    i_ = i + 0.5
    j_ = j + 0.5
    return [(i_,j_), (i_+1,j_), (i_+1,j_+1), (i_,j_+1)]

def demo(save=False):
    # fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(14,6))
    fig = plt.figure(figsize=(16, 10))
    gs = GridSpec(3,4)
    ax0 = fig.add_subplot(gs[:,0:3])
    ax1 = fig.add_subplot(gs[0,3])
    ax2 = fig.add_subplot(gs[1,3])
    ax3 = fig.add_subplot(gs[2,3])
    axes = [ax0, ax3]

    ax0.set_title("Simultaneous Probability Density of (X,Y) as Die Roll Pairs ", fontsize=16)
    ax1.set_title("Probability Density of X as Die Rolls", fontsize=16)
    ax2.set_title("Probability Density of Y as Die Rolls", fontsize=16)
    ax3.set_title("Probability Density of the Sum X + Y", fontsize=16)

    rects = []
    texts = []
    for i in range(6):
        row = []
        for j in range(6):
            vertices = create_vertices(i, j)
            rect = Polygon(vertices, alpha=0.5)
            ax0.add_patch(rect)
            t = ax0.text(i+1, j+1, "(%d,%d)" % (i+1,j+1), ha='center', va='center', fontsize=16)
            texts.append(t)
            row.append(rect)
        rects.append(row)

    vertices = create_vertices(0, 0)
    active_rect = Polygon(vertices, facecolor='red', alpha=0.5)
    ax0.add_patch(active_rect)

    ax0.set_xlim(0,7)
    ax0.set_ylim(0,7)
    ax0.set_aspect('equal')
    ax0.set_xlabel('X', fontsize=16)
    ax0.set_ylabel('Y', fontsize=16)
    ax0.grid(False)

    rects_array = np.array(rects)
    rects_list = list(rects_array.flatten())

    default_color = get_color(0)

    data_list = []
    num_frames = 0
    control_list = []
    for k in range(0,11):
        start = max(0, k-5)
        stop = min(k+1, 6)
        num_diag = stop - start
        for m in range(num_diag):
            control_list.append([m, start, stop, num_diag, k])
            data_list.append(k+2)

    control_array = np.array(control_list)
    num_frames = control_array.shape[0]
    data = np.array(data_list)

    die_rolls = np.arange(1,7)
    pdf1 = DiscretePdf(die_rolls, num_bins=6, num_frames=num_frames)
    pdf1.prepare(ax1)
    vertices = pdf1.get_nth_bar_vertices(0)
    pdf1_pnt = Polygon(vertices, facecolor='red', alpha=0.5)
    ax1.add_patch(pdf1_pnt)

    pdf2 = DiscretePdf(die_rolls, num_bins=6, num_frames=num_frames)
    pdf2.prepare(ax2)
    pdf2_pnt = Polygon(vertices, facecolor='red', alpha=0.5)
    ax2.add_patch(pdf2_pnt)

    pdf3 = DiscretePdf(data, num_bins=11, num_frames=num_frames)
    pdf3.prepare(ax3)

    for ax in [ax1, ax2]:
        ax.set_ylim(0, 0.3)

    vertices = pdf3.get_nth_bar_vertices(0)
    pdf3_bar = Polygon(vertices, facecolor='yellow')
    ax3.add_patch(pdf3_bar)
    pdf_bar_vertices = vertices
    pdf3_pnt = Polygon(pdf_bar_vertices, facecolor='red', alpha=0.5)
    ax3.add_patch(pdf3_pnt)

    ax3.set_ylim(0, 0.2)

    show_artists = [active_rect, pdf1_pnt, pdf2_pnt, pdf3_bar, pdf3_pnt]

    pdf_bar_pos = 0
    actual_n = 0
    pause = False

    def on_click(event):
        nonlocal pause
        pause ^= True

    fig.canvas.mpl_connect('button_press_event', on_click)

    def reset():
        if actual_n == 0:
            rects_array[5,5].set_facecolor(default_color)
            return rects_list + texts + show_artists
        else:
            return []

    def set_diag_color(n, color, alt_color=None):
        m, start, stop, num_diag, k = control_array[n]
        if alt_color is not None:
            if k % 2 == 1:
                color = alt_color
        for i in range(start, stop):
            j = k - i
            rects_array[i,j].set_facecolor(color)

    def move_active_rect(n):
        m, start, stop, num_diag, k = control_array[n]
        i = start + m
        j = k - i
        vertices = create_vertices(i, j)
        active_rect.set_xy(vertices)
        return i,j

    def pdf_pointer_vertices(vertices, n):
        m, num_diag = control_array[n][[0,3]]
        left_, bottom_ = vertices[0]
        right_, top_ = vertices[2]
        delta_height = (top_ - bottom_)/num_diag
        new_bottom = bottom_ + delta_height*m
        new_top = new_bottom + delta_height
        return [(left_, new_bottom), (right_, new_bottom), (right_, new_top), (left_, new_top)]

    def animate(n_):
        nonlocal pdf_bar_vertices, actual_n
        n = actual_n
        m, k = control_array[n][[0, -1]]
        if m == 0:
            if n > 0:
                set_diag_color(n-1, default_color)
            set_diag_color(n, 'cyan', alt_color='yellow')
            pdf_bar_vertices = pdf3.get_nth_bar_vertices(k)
            pdf3_bar.set_xy(pdf_bar_vertices)
            color = 'cyan' if k % 2 == 0 else 'yellow'
            pdf3_bar.set_facecolor(color)
        i, j = move_active_rect(n)
        vertices = pdf1.get_nth_bar_vertices(i)
        pdf1_pnt.set_xy(vertices)
        vertices = pdf2.get_nth_bar_vertices(j)
        pdf2_pnt.set_xy(vertices)
        vertices = pdf_pointer_vertices(pdf_bar_vertices, n)
        pdf3_pnt.set_xy(vertices)

        if not pause:
            actual_n += 1
            if actual_n == num_frames:
                actual_n = 0
                reset()

        return rects_list + texts + show_artists

    anim = animation.FuncAnimation(fig, animate, frames=num_frames, blit=True, init_func=reset, interval=500)

    fig.subplots_adjust(left=-0.05, right=0.95, bottom=0.05, top=0.95, wspace=-0.1, hspace=0.3)

    if save:
        anim.save("anim.mp4", writer="ffmpeg")

    plt.show()

def demo3d():
    pass



# coding: utf-8
"""
    SvdTutorial3D.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.patches import Polygon, Circle
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
import molass_legacy.KekLib.DebugPlot as dplt
from molass_legacy.KekLib.OurMatplotlib import get_color
from OurMatplotlib3D import plot_parallelepiped, plot_sphere
from ModeledData import ModeledData
from MatrixData import simple_plot_3d
from .SvdPrecision import plot_text

FONTSIZE = 200

class SvdTutorial3D(Dialog):
    def __init__(self, parent):
        self.parent = parent
        self.B = np.array([[1,0,0], [0,1,0], [0,0,1]]).T
        Dialog.__init__(self, parent, "SVD Tutorial (3D)", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        tk_set_icon_portable(self)

        cframe = Tk.Frame(body_frame)
        cframe.pack()

        self.fig = fig = plt.figure(figsize=(15,10))
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.mpl_canvas.mpl_connect('button_press_event', self.on_press)
        self.create_axes()
        self.draw()
        fig.tight_layout()
        self.mpl_canvas.draw()

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Animate", width=10, command=self.animate, state=Tk.DISABLED)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

    def create_axes(self):
        fig = self.fig
        gs = GridSpec(4,3)
        ax1 = fig.add_subplot(gs[1:3, 0], projection='3d')
        ax2 = fig.add_subplot(gs[0:2, 1], projection='3d')
        ax3 = fig.add_subplot(gs[1:3, 2], projection='3d')
        ax4 = fig.add_subplot(gs[2:4, 1], projection='3d')
        self.axes = [ax1, ax2, ax3, ax4]
        self.ax5 = fig.add_subplot(gs[3, 2])
        self.ax5.set_axis_off()
        plot_text(self.ax5, "Click anywhere to play", alpha=0.2, fontsize=30)

    def draw(self):
        self.generate_random_points()
        ax1, ax2, ax3, ax4 = self.axes
        self.draw1(ax1)
        self.draw2(ax2)
        self.draw3(ax3)
        self.draw4(ax4)

    def on_press(self, event):
        for ax in self.axes:
            ax.cla()
        self.draw()
        self.mpl_canvas.draw()

    def generate_random_points(self):
        a = np.random.rand()*2 - 1
        b = np.random.rand()*2 - 1
        c = np.random.rand()*2 - 1
        d = np.random.rand()*2 - 1
        e = np.random.rand()*2 - 1
        f = np.random.rand()*2 - 1
        g = np.random.rand()*2 - 1
        h = np.random.rand()*2 - 1
        i = np.random.rand()*2 - 1
        j = np.random.rand()*2 - 1

        self.points = [(a,b,c), (d,e,f), (h,i,j)]
        self.M = np.array(self.points).T
        U, s, VT = np.linalg.svd(self.M)
        self.U = U
        self.S = np.diag(s)
        self.V = VT.T

    def draw_points(self, ax, W):
        ax.plot([0], [0], [0], 'o', color='red')
        px = W[0,:]
        py = W[1,:]
        pz = W[2,:]
        ax.plot(px, py, pz, 'o')

    def draw_parallelepiped(self, ax, W, color=None, edgecolor='w'):
        p1 = W[:,0]
        p2 = W[:,1]
        p3 = W[:,2]
        plot_parallelepiped(ax, (0,0,0), [p1, p2, p3], color=color, edgecolor=edgecolor, alpha=0.5)

    def draw_sphere(self, ax):
        plot_sphere(ax, (0,0,0), 1, alpha=0.05, n=40, color='gray', edgecolor='gray')

    def draw1(self, ax):
        self.draw_points(ax, self.M)
        self.draw_parallelepiped(ax, self.B, color='pink', edgecolor='red')
        self.draw_parallelepiped(ax, self.M, color='cyan', edgecolor='blue')
        ax.set_xlim(-2, 2)
        ax.set_ylim(-2, 2)
        ax.set_zlim(-2, 2)

        # plot_text(ax, "$M$", fontsize=FONTSIZE)

    def draw2(self, ax):
        self.draw_sphere(ax)
        self.draw_points(ax, self.V)
        self.draw_parallelepiped(ax, self.B, color='pink', edgecolor='red')
        self.draw_parallelepiped(ax, self.V, color='cyan', edgecolor='blue')
        ax.set_xlim(-2, 2)
        ax.set_ylim(-2, 2)
        ax.set_zlim(-2, 2)

        # plot_text(ax, "$V^t$", fontsize=FONTSIZE)

    def draw3(self, ax):
        self.draw_points(ax, self.S)
        self.draw_parallelepiped(ax, self.B, color='pink', edgecolor='red')
        self.draw_parallelepiped(ax, self.S, color='cyan', edgecolor='blue')
        ax.set_xlim(-2, 2)
        ax.set_ylim(-2, 2)
        ax.set_zlim(-2, 2)

        # plot_text(ax, "$\Sigma$", fontsize=FONTSIZE)

    def draw4(self, ax):
        self.draw_sphere(ax)
        self.draw_points(ax, self.U)
        self.draw_parallelepiped(ax, self.B, color='pink', edgecolor='red')
        self.draw_parallelepiped(ax, self.U, color='cyan', edgecolor='blue')
        ax.set_xlim(-2, 2)
        ax.set_ylim(-2, 2)
        ax.set_zlim(-2, 2)

        # plot_text(ax, "$U$", fontsize=FONTSIZE)

    def animate(self):
        self.show_matrix_mapping()
        self.make_anim_VT()
        self.make_anim_S()
        self.make_anim_U()
        self.show_anim()

    def show_matrix_mapping(self):
        dplt.push()
        fig, axes = dplt.subplots(nrows=1, ncols=2, figsize=(12,6))
        ax1, ax2 = axes
        ax1.plot(0,0,'o')
        ax2.set_axis_off()
        fig.tight_layout()
        dplt.show()
        dplt.pop()

    def make_anim_VT(self):
        pass

    def make_anim_S(self):
        pass

    def make_anim_U(self):
        pass

    def show_anim(self):
        pass



"""
    FlowChangePrep.py

    Copyright (c) 2021-2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.stats import linregress
from molass_legacy.UV.PlainCurve import check_diehardness, flattenable

REMAINING_FLATTEN_MARGIN = 50   # required in 20211021 to modify the positions from integral
REMAINING_FLATTEN_PROPORTIONS = [0.05, 0.95]    # check at least for 20180617 to narrow, 20211021 to widen
MODIFY_REMAINING = True
REMAINING_RATIO_TEST = 0.1
REMAINING_RATIO_LIMIT = 0.15    # > 0.285 for 20170304, < 0.098 for 20181127

def flatten_remaining_peak(peak_region, y, debug=False):
    j1_, j2_ = peak_region.get_peak_ends(REMAINING_FLATTEN_PROPORTIONS)[1]
    j1 = max(0, j1_-REMAINING_FLATTEN_MARGIN)
    j2 = min(len(y)-1, j2_+REMAINING_FLATTEN_MARGIN)
    ret_y = y.copy()
    n = j2-j1
    w = np.arange(n)/(n-1)
    ret_y[j1:j2] = y[j1]*(1-w) + y[j2]*w
    if debug:
        import molass_legacy.KekLib.DebugPlot as plt

        with plt.Dp():
            x = np.arange(len(y))
            fig, ax = plt.subplots()
            ax.plot(x, y)
            ax.plot(x, ret_y, ":")
            ymin, ymax = ax.get_ylim()
            ax.set_ylim()
            for j in [j1_, j2_]:
                ax.plot([j, j], [ymin, ymax], ":", color="blue")
            for j in [j1, j2]:
                ax.plot([j, j], [ymin, ymax], ":", color="red")
            fig.tight_layout()
            plt.show()
            # exit()
    return ret_y

def get_easier_curve_y(a_curve, a_curve2, peak_region, logger, debug=False):
    x = a_curve2.x
    y = a_curve2.y

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt

        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("get_easier_curve_y entry")
            ax1.plot(a_curve.x, a_curve.y)
            ax2.plot(x, y)
            fig.tight_layout()
            plt.show()

    y_for_gy = y
    if MODIFY_REMAINING:
        y_modified = False
        diehard, max_k, ratio = check_diehardness(a_curve, a_curve2)
        if flattenable(ratio):
            # task: unify this and Uv.UvPreRecog.py
            pk_i = a_curve.primary_peak_i
            pk_y = np.average(a_curve2.y[pk_i-20:pk_i+21])
            remaining_ratio = pk_y/a_curve.primary_peak_y
            print("remaining_ratio=", remaining_ratio)
            if remaining_ratio > REMAINING_RATIO_TEST:
                outside = peak_region.get_outside()
                slope, intercept = linregress(x[outside], y[outside])[0:2]
                remaining_ratio = (pk_y - (slope*pk_i + intercept))/a_curve.primary_peak_y
                print("corrected remaining_ratio=", remaining_ratio)
                if remaining_ratio > REMAINING_RATIO_LIMIT:
                    # as in 20180219
                    y_for_gy = flatten_remaining_peak(peak_region, y)
                    y_modified = True
                    logger.info("remaining peak has been flattened due to remaining_ratio=%.3g > %.3g.", remaining_ratio, REMAINING_RATIO_LIMIT)
        else:
            # as in 20211021 where ratio == 1.54
            y_for_gy = np.zeros(len(y))
            y_modified = True
    else:
        y_modified = False

    first_peak, ret_y = peak_region.get_first_peak()
    last_peak = a_curve.peak_info[-1]

    if debug:
        print("diehard, max_k, ratio=", diehard, max_k, ratio)
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("get_easier_curve_y return")
            ax1.plot(a_curve.x, a_curve.y)
            ax2.plot(x, y)
            ax2.plot(x, y_for_gy)
            fig.tight_layout()
            plt.show()

    return y_modified, y_for_gy, first_peak, last_peak



"""
    Models.Stochastic.PairedAnalysis.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from importlib import reload
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder

def paired_analysis_main(info_list, exec_spec):
    from molass_legacy.Models.Stochastic.PairedAnalysisSpecs import get_components_to_use
    from molass_legacy.Models.Stochastic.MomentUtils import compute_egh_moments
    num_components  = exec_spec["num_components"]
    componets_to_use = get_components_to_use(exec_spec)
    print("paired_analysis_impl", len(info_list))

    def ssdc_analysis():
        from SSDC.SsdcAnalysis import SsdcAnalysis
        parent = plt.get_parent()
        dialog = SsdcAnalysis(parent, info_list, num_components=num_components)
        dialog.show()

    def ssdc_analysis_sdm():
        import SSDC.SsdcAnalysisSdm
        reload(SSDC.SsdcAnalysisSdm)
        from SSDC.SsdcAnalysisSdm import spike
        spike(info_list, num_components, exec_spec)
    
    extra_button_specs = [
                ("SSDC Analysis", ssdc_analysis),
                ("SSDC Analysis (SDM)", ssdc_analysis_sdm),
                ]

    with plt.Dp(button_spec=["OK", "Cancel"],
                extra_button_specs=extra_button_specs):
        fig, axes = plt.subplots(ncols=2, figsize=(12,5))

        average_sigma_list = []
        for k, (in_folder, lrfsrc) in enumerate(info_list):
            ax = axes[k]
            ax.set_title(get_in_folder(in_folder))
            select = componets_to_use[k]
            rgs, trs, orig_props, peak_rgs, peak_trs, props, indeces, qualities = lrfsrc.compute_rgs(want_num_components=num_components, select=select, debug=False)
            x = lrfsrc.xr_x
            y = lrfsrc.xr_y
            model = lrfsrc.model
            ax.plot(x, y)
            cy_list = []
            print([k], "indeces=", indeces)
            peaks = lrfsrc.xr_peaks[indeces]
            moments_list = compute_egh_moments(peaks)
            print([k], "props=", props)
            print([k], "moments_list=", moments_list)
            average_sigma = np.sum(props * np.sqrt(np.array(moments_list)[:,1]))
            average_sigma_list.append(average_sigma)
            for k, params in enumerate(peaks):
                cy = model(x, params)
                ax.plot(x, cy, ":", label="$R_g$=%.3g" % peak_rgs[k])
            ax.legend()
        
        print("sigma ratio=", average_sigma_list[0]/average_sigma_list[1])

        fig.tight_layout()
        plt.show()


"""
    Selective.ModelSelectFrame.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""

from molass_legacy.KekLib.OurTkinter import Tk, ttk
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy._MOLASS.Version import is_developing_version
from molass_legacy.GuiParts.ColumnTypeSelector import ColumnTypeSelector

enable_edm_model = get_setting('enable_edm_model')
if enable_edm_model:
    # MODEL_NAMES = ['EMG', 'EGH', 'EDM', 'STC', 'SDM'] # SDM is not used in the current version
    MODEL_NAMES = ['EMG', 'EGH', 'EDM', 'STC']
else:
    MODEL_NAMES = ['EMG', 'EGH']

class ModelSelectFrame(Tk.LabelFrame):
    def __init__(self, parent, editor, **kwargs):
        Tk.LabelFrame.__init__(self, parent, **kwargs)

        self.editor = editor
        select_frame = Tk.Frame(self)
        select_frame.pack(padx=10, pady=20)

        sf_row = 0
        label = Tk.Label(select_frame, text="Elution Curve Model: ")
        label.grid(row=sf_row, column=0, sticky=Tk.E)
        self.model_name  = Tk.StringVar()
        self.model_name.set(MODEL_NAMES[editor.current])
        w = ttk.Combobox(select_frame, textvariable=self.model_name, values=MODEL_NAMES,
                         width=10, justify=Tk.CENTER, state="readonly")
        w.grid(row=sf_row, column=1, sticky=Tk.W, padx=5, pady=10)
        w.bind("<<ComboboxSelected>>", self.change_model)
        self.model_select_box = w

        sf_row += 1
        self.use_column_constraints = Tk.IntVar()
        self.use_column_constraints.set(0)
        self.ucc_check_button = Tk.Checkbutton(select_frame, text="Use Column Constraints", variable=self.use_column_constraints)
        self.ucc_check_button.grid(row=sf_row, column=0, columnspan=2, sticky=Tk.W, padx=5, pady=10)
        self.use_column_constraints.trace_add("write", self.use_column_constraints_tracer)

        sf_row += 1
        label = Tk.Label(select_frame, text="SEC Column Type: ")
        label.grid(row=sf_row, column=0, sticky=Tk.E)
        self.selector = ColumnTypeSelector(select_frame)
        self.selector.grid(row=sf_row, column=1, sticky=Tk.W, padx=5, pady=0)

        sf_row += 1
        button = Tk.Button(select_frame, text="PDS Simulation", command=self.show_psd_simulation)
        button.grid(row=sf_row, column=0, columnspan=2, pady=30)

        if is_developing_version():
            button = Tk.Button(select_frame, text="Devel Test", command=self.do_devel_test)
            button.grid(row=sf_row, column=2, columnspan=2, padx=20)

    def change_model(self, *args):
        self.editor.change_model(self.model_name.get())

    def get_model_name(self):
        return self.model_name.get()

    def use_column_constraints_tracer(self, *args):
        state = Tk.DISABLED if self.use_column_constraints.get() == 0 else Tk.NORMAL
        self.selector.config(state=state)

    def update_selection_state(self):
        if self.editor.get_current_model().is_traditional():
            state = Tk.DISABLED
            self.use_column_constraints.set(0)
        else:
            state = Tk.NORMAL        
        self.ucc_check_button.config(state=state)

    def show_psd_simulation(self, devel=True):
        import molass_legacy.KekLib.CustomMessageBox as MessageBox
        yn = MessageBox.askyesno("Simulation Start Confirmation",
            'It may take several minutes to prepare for the simulation.\n'
            'Are you sure to proceed?"\n'
            '(Although, it will not be blocked since it will start as a separate process)"\n'
            ,
            parent=self.editor.parent
            )

        if not yn:
            return

        if devel:
            from importlib import reload
            import Simulative.PsdSimulation
            reload(Simulative.PsdSimulation)
        from Simulative.PsdSimulation import pds_simulation_impl

        pds_simulation_impl(self.editor)

    def do_devel_test(self):
        from importlib import reload
        import Selective.DevelTest
        reload(Selective.DevelTest)
        from Selective.DevelTest import devel_test_impl
        devel_test_impl(self.editor)



"""
    Simulative.SimulationBridge.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import os
from threading import Thread
from molass_legacy.KekLib.ChangeableLogger import Logger
from molass_legacy.KekLib.OurTkinter import Tk, ttk, Dialog
from molass_legacy._MOLASS.SerialSettings import load_settings, get_setting

MAXNUM_STEPS = 10

class PreparationProgress(Dialog):
    def __init__(self, parent, lrf_src):
        load_settings(lrf_src.run_setting_file)
        analysis_folder = get_setting('analysis_folder')
        log_file = os.path.join(analysis_folder, "simulation.log")
        self.logger = Logger(log_file)  # The logging module is thread-safe
        self.parent = parent
        self.lrf_src = lrf_src
        self.canceled = False
        self.ready = False
        Dialog.__init__(self, parent, "Preparation Progress", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        label = Tk.Label(body_frame, text="Preparation Progress")
        label.pack(padx=10, pady=10)
        self.pbar = ttk.Progressbar(body_frame, orient ="horizontal", length=400, mode="determinate")
        self.pbar["maximum"] = MAXNUM_STEPS
        self.pbar["value"] = 0
        self.pbar.pack(padx=10, pady=10)

        self.parent.after(100, self.start_preparation)

    def buttonbox( self ):
        box = Tk.Frame(self)
        box.pack()
        w = Tk.Button(box, text="Cancel", width=10, command=lambda: self.user_cancel(ask=True) )
        w.pack(side=Tk.LEFT, padx=5, pady=5)

    def apply(self):
        self.applied = True

    def user_cancel(self, ask=False):
        if ask:
            if not self.ask_cancel():
                return
        self.prepare_thread.terminate()
        self.cancel()

    def ask_cancel(self):
        from tkinter import messagebox
        ret = messagebox.askokcancel("Cancel", "Are you sure you want to cancel?")
        if ret:
            self.canceled = True

    def update_progress(self):
        if not self.ready and not self.canceled:
            self.parent.after(200, self.update_progress)
        else:
            if self.canceled:
                pass
            else:
                self.prepare_thread.join()
                self.ok()

    def start_preparation(self):
        self.prepare_thread = Thread(target=self.prepare)
        self.prepare_thread.start()
        self.update_progress()

    def prepare(self, use_study=True):
        if use_study:
            from importlib import reload
            import Models.Stochastic.MomentsStudy
            reload(Models.Stochastic.MomentsStudy)
            from molass_legacy.Models.Stochastic.MomentsStudy import moments_study_impl
            self.guess_info = moments_study_impl(self.lrf_src, return_rgs=True, progress_cb=self.progress_cb)
        else:
            self.guess_info = self.lrf_src.guess_lnpore_params(return_rgs=True, progress_cb=self.progress_cb)
        self.ready = True

    def progress_cb(self, progress):
        self.pbar["value"] = progress

def simulation_bridge_impl(parent, lrf_src):
    prep = PreparationProgress(parent, lrf_src)
    prep.show()
    if not prep.canceled:
        parent.quit()

    from importlib import reload
    import Simulative.SimulationCushion
    reload(Simulative.SimulationCushion)
    from Simulative.SimulationCushion import demo_cushion

    demo_cushion(lrf_src, prep.guess_info, parent=parent)

def simulation_bridge(lrf_src):
    from molass_legacy.KekLib.TkUtils import get_tk_root
    print('simulation_main')
    root = get_tk_root()
    root.after(1000, lambda: simulation_bridge_impl(root, lrf_src))
    root.mainloop()


# coding: utf-8
"""

    ファイル名：   PilatusUtilsOldStyle.py

    処理内容：

       Pilatus 画像データ関連の共通処理

    Copyright (c) 2015-2020, 2025, SAXS Team, KEK-PF
"""
import os
import sys
import glob
import re
import pandas as pd
from DebugQueue             import debug_queue_get

separator_line  = re.compile( '^\s*$' )
exec_date_re    = re.compile( '^#Execution date:\s+(.+)$' )
prefix_re       = re.compile( '^#\s+Pilatus1 fileprefix:\s+(\w+)' )
num_cycles_re   = re.compile( '^#\s+Pilatus1 Number of cycles:\s+(\d+)' )
num_changes_re  = re.compile( '^#\s+Pilatus1 Detector position:\s+Change\s+(\d+)' )
position_re     = re.compile( '^#\s+Pilatus1 Detector position(\(original\)|\d+):\s+([+-]?\d+\.?\d*),([+-]?\d+\.?\d*)' )
wave_length_re  = re.compile( r'^#\s+Wavelength:\starget=(\d+\.\d+)' )
energy_re       = re.compile( r'^#\s+Energy:\s+target=(\d+)' )

image_ext_re = re.compile( '\.(tif|cbf)$', flags=re.I )

DEBUG = False

def get_ext( file ):
    m = image_ext_re.search( file )
    if m:
        return m.group( 1 )
    return None

def regex_glob( regex ):
    global globbed_extension_
    # return filter( lambda x: regex.search( x ), glob.glob( '*.*' ) )
    files = []
    for file in glob.glob( '*.*' ):
        m = regex.search( file )
        if m:
            globbed_extension_ = m.group( 1 )
            files.append( file )
    return files

def get_data_info( folder_info, log_file, mask_file,
                    adj_folder, syn_folder, pilatus_counter, counter_id,
                    log_file_path,
                    sample_complete=False, for_test_data=False,
                    logger=None ):

    in_folder = folder_info.path
    dict_info = get_dict_info( in_folder, adj_folder, syn_folder, pilatus_counter, counter_id )

    if debug_queue_get() == __name__ + '.die()': die()  # die(): undefined function

    prefix_info_list, text_dict = get_prefix_info_list( in_folder, log_file_path )
    prefix_dict = {}
    for info in prefix_info_list:
        # print(info[0])
        prefix_dict[info[0]] = info

    image_data_info = []

    for info in prefix_info_list:
        fkey_type_is_old = get_a_prefix_record_old_style( counter_id, dict_info, sample_complete, info, image_data_info )

        if not fkey_type_is_old:
            get_a_prefix_record_new_style( counter_id, dict_info, sample_complete, info, image_data_info )

    data_array = make_data_array( image_data_info, prefix_dict )

    if DEBUG:
        for i, rec in enumerate( prefix_info_list ):
            print( 'prefix_info_list[%d]=' % i, rec )
        if len(image_data_info) > 0:
            print( 'image_data_info[ 0 ]=', image_data_info[ 0 ] )
            print( 'image_data_info[ -1 ]=', image_data_info[ -1 ] )
        else:
            print( 'image_data_info is empty' )

    if for_test_data:
        return data_array, text_dict
    else:
        return log_file, mask_file, data_array, pilatus_counter

def get_prefix_info_list( in_folder, log_file_path ):
    prefix = None
    measurement_text = ''
    text_dict = {}
    exec_date = None

    fh = open( log_file_path )

    prefix_info_list = []

    for line in fh.readlines():
        # if DEBUG: print( line )

        if separator_line.match( line ):
            if prefix != '':
                text_dict[ prefix ] = measurement_text
            measurement_text = ''
            exec_date = None
        else:
            measurement_text += line

        me = exec_date_re.match( line )
        if me:
            exec_date = me.group(1)

        m0 = prefix_re.match( line )
        if m0:
            if prefix is not None:
                prefix_info_list.append( [ prefix, num_cycles, num_changes, pos_array, wave_lengths, energies, exec_date ] )

            prefix = m0.group( 1 )
            if prefix[-1] == '_':
                prefix = prefix[:-1]
            if DEBUG: print( 'prefix=', prefix )
            # pos_array = [ prefix ]

            pos_info_counter = 0
            num_cycles  = 0
            num_changes = 0
            pos_array = []
            wave_lengths = []
            energies = []
            continue

        m1 = num_cycles_re.match( line )
        if m1:
            num_cycles = int( m1.group( 1 ) )
            if DEBUG: print( 'num_cycles=', num_cycles )

        m2 = num_changes_re.match( line )
        if m2:
            # print line
            num_changes = int( m2.group( 1 ) )
            if DEBUG: print( 'num_changes=', num_changes )
            continue

        m3 = position_re.match( line )
        if m3:
            position = m3.group( 1 )
            x = m3.group( 2 )
            y = m3.group( 3 )
            if DEBUG: print( "\tposition_re.match: %s\t%s,%s" % (position, x, y) )
            pos_array.append( [ x, y ] )
            continue

        m4 = wave_length_re.match( line )
        if m4:
            wave_lengths.append( '_' + str( int( float( m4.group(1) ) * 1000 ) ) + 'A' )

        m5 = energy_re.match( line )
        if m5:
            energies.append( '_' + m5.group(1) + 'eV' )

    if prefix is not None:
        prefix_info_list.append( [ prefix, num_cycles, num_changes, pos_array, wave_lengths, energies, exec_date ] )

    if measurement_text != '':
        text_dict[ prefix ] = measurement_text

    fh.close()

    return prefix_info_list, text_dict

def get_dict_info( in_folder, adj_folder, syn_folder, pilatus_counter, counter_id ):
    orig_folder = os.getcwd()

    counter_dict = pilatus_counter.get_counter_dict( counter_id )
    if False:
        print( 'counter_id=', counter_id )
        for k, v in counter_dict.items():
            print( k, v )

    global globbed_extension_
    globbed_extension_ = None
    org_file_dict = {}
    if in_folder and os.path.exists( in_folder ):
        os.chdir( in_folder )
        for file in regex_glob( image_ext_re ):
            fkey = re.sub( r'_\d+\.\w+$', '', file, flags=re.I )
            # print( 'fkey=', fkey )
            org_file_dict[ fkey ] = file

    # if DEBUG: print( 'org_file_dict=', org_file_dict )
    # print( 'globbed_extension_=', globbed_extension_ )
    if globbed_extension_:
        restricted_image_ext_re = re.compile( '\.(' + globbed_extension_ + ')$' )
    else:
        restricted_image_ext_re = image_ext_re

    adj_file_dict = {}
    if adj_folder and os.path.exists( adj_folder ):
        os.chdir( adj_folder )
        for file in regex_glob( restricted_image_ext_re ):
            fkey = re.sub( r'_[^_]+\.\w+$', '', file, flags=re.I )
            adj_file_dict[ fkey ] = file

    # print( 'adj_file_dict=', adj_file_dict )

    syn_file_dict = {}
    if syn_folder and os.path.exists( syn_folder ):
        os.chdir( syn_folder )
        for file in regex_glob( restricted_image_ext_re ):
            fkey = re.sub( r'_[^_]+\.\w+$', '', file, flags=re.I )
            syn_file_dict[ fkey ] = file

    # print( 'syn_file_dict=', syn_file_dict )

    os.chdir( orig_folder )

    dict_info = [ counter_dict, org_file_dict, adj_file_dict, syn_file_dict ]
    return dict_info

def get_a_prefix_record_old_style( counter_id, dict_info, sample_complete, info, image_data_info ):
    old_result = True

    counter_dict, org_file_dict, adj_file_dict, syn_file_dict   = dict_info
    prefix, num_cycles, num_changes, pos_array, wave_lengths, energies, exec_date  = info

    if DEBUG:
        print( 'get_a_prefix_record_old_style' )
        print( 'prefix=', prefix, 'num_cycles=', num_cycles, 'num_changes=', num_changes, 'counter_id=', counter_id )

    info_array = []
    for i in range( num_changes ):
        fkey = '%s_%d' % ( prefix, i )
        if DEBUG: print( [i], 'fkey=', fkey )
        if i == 0:
            if num_cycles == 1:
                new_prefix_ = prefix
            else:
                new_prefix_ = '%s_%d' % ( prefix, i )
            new_fkey = '%s_d%d' % ( new_prefix_, i )

            orig_file_count = counter_dict.get( fkey, 0 )
            orig_file_count_new = counter_dict.get( new_fkey, 0 )
            if DEBUG:
                print( 'orig_file_count=', orig_file_count )
                print( 'orig_file_count_new=', orig_file_count_new )
            if orig_file_count == 0 or orig_file_count_new > 0:
                old_result = False
                if DEBUG: print( 'old_result is set to False' )
                break
            else:
                ratio = 1.0
        else:
            if counter_id == 'None':
                ratio = 1.0
            else:
                if orig_file_count > 0:
                    ratio = counter_dict.get( fkey ) / orig_file_count
                else:
                    ratio = None
        org_file = org_file_dict.get( fkey )
        syn_file = syn_file_dict.get( prefix )
        if not syn_file:
            # prefix_syn.tif がなければ prefix_{i}_syn.tif を設定する。
            syn_file = syn_file_dict.get( fkey )
        if org_file and syn_file:
            org_ext = get_ext( org_file )
            syn_ext = get_ext( syn_file )
            # TODO: avoid this failure
            assert( syn_ext == org_ext )

        info_array.append( [ org_file, pos_array[i], ratio, adj_file_dict.get( fkey ), syn_file ] )

    if len( info_array ):
        # print( 'sample_complete=', sample_complete, ', len( info_array )=', len( info_array ) )
        if not sample_complete or len( info_array ) == num_changes:
            # すなわち、 sample_complete == True の場合は num_changes に見たなければ
            # 採集しない。
            image_data_info.append( [ prefix, info_array ] )

    return old_result

def get_a_prefix_record_new_style( counter_id, dict_info, sample_complete, info, image_data_info ):
    counter_dict, org_file_dict, adj_file_dict, syn_file_dict   = dict_info
    prefix, num_cycles, num_changes, pos_array, wave_lengths, energies, exec_date  = info

    if DEBUG:
        print( 'fkey_type_is_new: prefix=', prefix, 'num_cycles=', num_cycles, 'counter_id=', counter_id )
        print( 'wave_lengths=', wave_lengths )
        print( 'energies=', energies )

    for suppress_limit in [ 2, 1 ]:
        # 命名規則としては suppress_limit == 2 であるが、
        # 20180221データのように、途中で止めた場合、suppress_limit == 1 での処理が必要。

        wave_length_list    = [ '' ] if len(wave_lengths) < suppress_limit else wave_lengths
        energy_list         = [ '' ] if len(energies) < suppress_limit else energies

        num_key_oks     = 0
        num_key_errors  = 0
        for wave_length in wave_length_list:
            for energy in energy_list:
                for j in range( num_cycles ):
                    if num_cycles == 1:
                        cycle = ''
                    else:
                        cycle = '_%d' % ( j )

                    info_array = []
                    for k in range( num_changes ):
                        new_prefix_ = prefix + wave_length + energy + cycle
                        fkey = '%s_d%d' % ( new_prefix_, k )
                        if DEBUG: print( [k], 'fkey=', fkey )
                        if k == 0:
                            orig_file_count = counter_dict.get( fkey, 0 )
                            if orig_file_count == 0:
                                if DEBUG: print( 'no count for fkey=', fkey )
                                num_key_errors += 1
                                break
                            else:
                                fkey_type_is_new = True
                                ratio = 1.0
                                num_key_oks += 1
                        else:
                            if counter_id == 'None':
                                ratio = 1.0
                            else:
                                if orig_file_count > 0:
                                    changed_file_count = counter_dict.get( fkey, 0 )
                                    if changed_file_count > 0:
                                        ratio = counter_dict.get( fkey ) / orig_file_count
                                        # print( 'ratio=', ratio )
                                    else:
                                        # stopped?
                                        continue
                                else:
                                    assert( False )
                        syn_file = syn_file_dict.get( new_prefix_ )
                        if not syn_file:
                            # TODO: 必要？
                            pass

                        info_array.append( [ org_file_dict.get( fkey ), pos_array[k], ratio, adj_file_dict.get( fkey ), syn_file ] )

                    if len( info_array ):
                        if not sample_complete or len( info_array ) == num_changes:
                            # すなわち、 sample_complete == True の場合は num_changes に満たなければ
                            # 採集しない。
                            image_data_info.append( [ new_prefix_, info_array, prefix ] )
                            # [ new_prefix_, info_array, prefix ] will be interpreted
                            # [ prefix, info_array, dict_key ] later in make_data_array below

        if num_key_oks > 0:
            break

def make_data_array( image_data_info, prefix_dict ):
    data_array = []

    for info in image_data_info:
        if len(info) == 2:
            prefix, info_rec = info
            dict_key = prefix
        elif len(info) == 3:
            prefix, info_rec, dict_key = info
        else:
            assert False
        # print('dict_key=', dict_key)

        is_lacking = False
        for rec in info_rec:
            if rec[0] is None:
                is_lacking = True

        if is_lacking:
            # いずれかの画像がないデータは無視する。
            continue

        info = prefix_dict.get(dict_key)
        num_changes = info[2]
        # print( prefix, info_rec )
        data_array.append( [ prefix, info_rec, num_changes] )
        # self.data_array は、表示された Table オブジェクトのセル（表示値）と
        # 単純な対応関係を持つように構築するが、
        # その平坦な構造上、prefix ごとのグルーピングが失われるため、
        # i によってそのグルーピングを把握できるようにしておく。
        """
                [ AgBh002,      [ AgBh002_0_00000.tif,  ['0.70000','-0.35000'], 1.0, None, None ],
                                [ AgBh002_1_00000.tif,  ['5', '3'],             1.0, AgBh002_1_adj.tif, AgBh002_1_syn.tif ],
                                [ AgBh002_2_00000.tif,  ['-5', '-3'],           1.0, AgBh002_2_adj.tif, AgBh002_2_syn.tif ],
                ]
        """

    return data_array



# coding: utf-8
"""
    NormalVectorDiff.py

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF
"""

import numpy                as np
from scipy.interpolate      import UnivariateSpline

VERY_SMALL_VALUE = 1e-6

class NormalVectorDiff:
    def __init__(self, x, y1, spline1=None):
        self.x  = x
        self.y1 = y1
        if spline1 is None:
            spline1 = UnivariateSpline(x, y1, s=0)
        d1 = spline1.derivative(1)(x)
        self.ratio = 1/np.sqrt(1 + d1**2)

    def diff(self, y2):
        dy = np.abs( y2 - self.y1 )
        return dy * self.ratio


"""
    FoulingBaseline.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
import molass_legacy.KekLib.DebugPlot as plt
from .LinearBaseline import USE_END_PARAMS
if not USE_END_PARAMS:
    from .Constants import SLOPE_SCALE

OPTIMIZE_AB = True

class FoulingBaseline:
    def __init__(self, x=None, y=None, debug=False):
        if debug:
            from importlib import reload
            import Baseline.Baseline
            reload(Baseline.Baseline)            
        from .Baseline import compute_baseline

        if x is None:
            return

        yb, (a, b, r) = compute_baseline(y, x=x, integral=True, return_params=True, debug=debug)
        self.x = x
        self.y = y
        self.yb = yb
        self.x1 = x[0]
        self.x2 = x[-1]

        if OPTIMIZE_AB:
            y_ = y - yb

            def func(p):
                a_, b_ = p
                params = [a_*SLOPE_SCALE, b_, r]
                yb_ = self.__call__(x, params, y_, [y_])
                return np.sum((yb_ - yb)**2)

            ret = minimize(func, (a,b))
            a_, b_ = ret.x
        else:
            a_, b_ = a, b

        self.params = [a_*SLOPE_SCALE, b_, r]

        if debug:
            yb_ = self.__call__(x, self.params, y_)
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("IntegralBaseline.__init__")
                ax.plot(x, y)
                ax.plot(x, yb, ":")
                ax.plot(x, yb_, ":")
                fig.tight_layout()
                plt.show()

        y1, y2 = [a_*px+ b_ for px in x[[0,-1]]]
        self.end_params = [y1, y2, r]

    def __call__(self, x, params, y_, cy_list):

        if USE_END_PARAMS:
            y1, y2, r = params
            k = (y2 - y1)/(self.x2 - self.x1)
            y_linear = y1 + k*(x - self.x1)
        else:
            a, b, r = params
            y_linear = x*a/SLOPE_SCALE + b
            y_integral = np.zeros(len(x))
            for cy in cy_list:
                y_integral += r*np.cumsum(cy)
        return y_linear + y_integral

    def get_baseplane(self, D, j1, j2, debug=True):
        bz = D[:,j2] - D[:,j1]
        scale = 1/(self.yb[j2] - self.yb[j1])
        B = scale * bz[:,np.newaxis] @ self.yb[np.newaxis,:]
        if debug:
            from MatrixData import simple_plot_3d
            with plt.Dp():
                fig = plt.figure(figsize=(18,5))
                ax1 = fig.add_subplot(131)
                ax2 = fig.add_subplot(132)
                ax3 = fig.add_subplot(133, projection="3d")
                ax1.plot(self.x, self.y)
                ax2.plot(self.yb)
                simple_plot_3d(ax3, B)
                fig.tight_layout()
                plt.show()
        return B



"""
    stack overflow
    Getting standard errors on fitted parameters using the optimize.leastsq method in python
"""
import random
import numpy as np
from scipy import optimize

def f( x, p0, p1, p2):
    return p0*x + 0.4*np.sin(p1*x) + p2

def ff(x, p):
    return f(x, *p)

# These are the true parameters
p0 = 1.0
p1 = 40
p2 = 2.0

# These are initial guesses for fits:
pstart = [
    p0 + random.random(),
    p1 + 5.*random.random(), 
    p2 + random.random()
]

# %matplotlib inline
import matplotlib.pyplot as plt
xvals = np.linspace(0., 1, 120)
yvals = f(xvals, p0, p1, p2)

# Generate data with a bit of randomness
xdata = np.array(xvals)
np.random.seed(42)
err_stdev = 0.2
yvals_err =  np.random.normal(0., err_stdev, len(xdata))
ydata = f(xdata, p0, p1, p2) + yvals_err

plt.plot(xvals, yvals)
plt.plot(xdata, ydata, 'o', mfc='None')
plt.show()

def fit_leastsq(p0, datax, datay, function):

    errfunc = lambda p, x, y: function(x,p) - y

    pfit, pcov, infodict, errmsg, success = \
        optimize.leastsq(errfunc, p0, args=(datax, datay), \
                          full_output=1, epsfcn=0.0001)

    if (len(datay) > len(p0)) and pcov is not None:
        s_sq = (errfunc(pfit, datax, datay)**2).sum()/(len(datay)-len(p0))
        pcov = pcov * s_sq
    else:
        pcov = np.inf

    error = [] 
    for i in range(len(pfit)):
        try:
          error.append(np.absolute(pcov[i][i])**0.5)
        except:
          error.append( 0.00 )
    pfit_leastsq = pfit
    perr_leastsq = np.array(error) 
    return pfit_leastsq, perr_leastsq 

pfit, perr = fit_leastsq(pstart, xdata, ydata, ff)

print("\nFit paramters and parameter errors from lestsq method :")
print("pfit = ", pfit)
print("perr = ", perr)



# coding: utf-8
"""
    Rgg.BayesPySpike.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
from time import time
import asyncio
import numpy as np
from matplotlib.gridspec import GridSpec
import molass_legacy.KekLib.DebugPlot as plt
from bayespy.nodes import Dirichlet, Categorical
from bayespy.nodes import Gaussian, Wishart
from bayespy.nodes import Mixture
from bayespy.inference import VB
import bayespy.plot as bpplt
from RgProcess.RgCurve import make_availability_slices
from .SpikeData import generate_demo_data
from .RggUtils import convert_to_probabilitic_data, plot_histogram_2d

BAYESPY_PLOT = False

def spike_demo():
    x, y, rg, num_components = generate_demo_data()
    spike_demo_impl([(x, y, rg, None)], num_components=num_components)

def spike_demo_impl(data_list, num_components=None):

    num_datasets = len(data_list)

    if BAYESPY_PLOT:
        pass
    else:
        plt.push()
        fig = plt.figure(figsize=(21,7))
        gs = GridSpec(num_datasets, 3)

    for k, (x, y, rg, X) in enumerate(data_list):

        if BAYESPY_PLOT:
            pass
        else:
            ax1 = fig.add_subplot(gs[k,0])
            ax2 = fig.add_subplot(gs[k,1], projection='3d')
            ax3 = fig.add_subplot(gs[k,2])

            axt = ax1.twinx()
            axt.grid(False)

            # ax.plot(x, y)
            # axt.plot(x, rg, color='C1')

        max_y = np.max(y)

        X_list = []
        slices, states = make_availability_slices(y, max_y=max_y, min_ratio=0.03)
        for slice_, state in zip(slices, states):
            x_ = x[slice_]
            y_ = y[slice_]
            rg_ = rg[slice_]
            if state:
                X_list.append(convert_to_probabilitic_data(x_, y_, rg_, max_y=max_y))

            if BAYESPY_PLOT:
                pass
            else:
                linestyle = '-' if state else ':'
                ax1.plot(x_, y_, linestyle, color='C0')
                axt.plot(x_, rg_, linestyle, color='C1')
                if state:
                    if True:
                        plot_histogram_2d(ax2, x_, y_, rg_, max_y)
                    else:
                        b_ = np.zeros(len(x_))
                        ax2.bar3d(x_, rg_, b_, 0.5, 0.05, y_/max_y*100, shade=True, edgecolor='green')
                    ax2.set_ylim(10, 60)

                ymin, ymax = axt.get_ylim()
                axt.set_ylim(0, ymax*1.2)

        X = np.concatenate(X_list)
        print("X.shape=", X.shape)

        N, D = X.shape
        K = 3
        alpha = Dirichlet(1e-5*np.ones(K),
                          name='alpha')
        Z = Categorical(alpha,
                        plates=(N,),
                        name='z')

        mu = Gaussian(np.zeros(D), 1e-5*np.identity(D),
                      plates=(K,),
                      name='mu')
        Lambda = Wishart(D, 1e-5*np.identity(D),
                         plates=(K,),
                         name='Lambda')

        Y = Mixture(Z, Gaussian, mu, Lambda,
                    name='Y')
        Z.initialize_from_random()
        Q = VB(Y, mu, Lambda, Z, alpha)
        Y.observe(X)
        Q.update(repeat=1000)
        # print("dir(Y)", dir(Y))
        get_component_params(Y)

        if BAYESPY_PLOT:
            bpplt.gaussian_mixture_2d(Y, alpha=alpha, scale=2)
            bpplt.pyplot.show()

    if BAYESPY_PLOT:
        pass
    else:
        fig.tight_layout()
        cont = plt.show()
        plt.pop()

def get_component_params(X):
    """
    from
    bayespy.plot.py - gaussian_mixture_2d
    """
    import scipy
    from bayespy.inference.vmp.nodes.gaussian import (GaussianMoments,
                                                      GaussianWishartMoments)

    mu_Lambda = X._ensure_moments(X.parents[1], GaussianWishartMoments)

    (mu, _, Lambda, _) = mu_Lambda.get_moments()
    mu = np.linalg.solve(Lambda, mu)

    if len(mu_Lambda.plates) != 1:
        raise NotImplementedError("Not yet implemented for more plates")

    K = mu_Lambda.plates[0]

    width = np.zeros(K)
    height = np.zeros(K)
    angle = np.zeros(K)

    for k in range(K):
        m = mu[k]
        L = Lambda[k]
        (u, W) = scipy.linalg.eigh(L)
        u[0] = np.sqrt(1/u[0])
        u[1] = np.sqrt(1/u[1])
        width[k] = 2*u[0]
        height[k] = 2*u[1]
        angle[k] = np.arctan(W[0,1] / W[0,0])

    angle = 180 * angle / np.pi
    mode_height = 1 / (width * height)
    print("mu=", mu)
    print("width=", width)
    print("height=", height)
    print("angle=", angle)

    params = sorted([(m, w, h) for m, w, a in zip(mu[:,0], width, height)], key=lambda x: x[0])
    print("params=", params)

def spike_demo_real(in_folder, num_components=3, correction=True, use_peaks=False):
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from RgProcess.RgCurve import RgCurve

    sp = StandardProcedure()
    sp.load(in_folder, debug=False)
    if correction:
        sd = sp.get_corrected_sd()
    else:
        sd = sp.get_sd()
    D, E, qv, xr_curve = sd.get_xr_data_separate_ly()
    rgc = RgCurve(qv, xr_curve, D, E)
    rgc.proof_plot()
    X = rgc.get_probabilistic_data()
    spike_demo_impl([[xr_curve.x, xr_curve.y, None, X]], num_components=3, use_peaks=use_peaks)

async def get_content(n):
    await asyncio.sleep(3)
    return n + 1

async def f(n):
    content = await get_content(n)
    return content

def coroutine_spike():
    loop = asyncio.get_event_loop()
    v = loop.run_until_complete(f(1))
    print(v)



import pandas as pd
import math
import os
import matplotlib.pyplot as plt
plt.style.use('seaborn-whitegrid')
import numpy as np
from PIL import Image
import glob
import moviepy.editor as mp
from datetime import datetime
import time

# BASIC FUNCTIONS

# Sqr and Cube functions:

def Sqr(x):
    return (x * x) 

def Cube(x):
    return ((x) * (x) * (x))

  
# Randomness functions: 

def RandR():
    global randSeedP
    randSeedP = (randSeedP * IMUL + IADD) & MASK
    return (randSeedP * SCALE)

def VRand(p):
    s: float
    s = 2. * math.pi * RandR()
    p[0] = math.cos(s)
    p[1] = math.sin(s)
    return p


# Toroidal functions:
def VWrapAll(v):
    if v[0] >= 0.5 * region[0]:
        v[0] -= region[0]
    elif v[0] < -0.5 * region[0]:
        v[0] += region[0]
        
    if v[1] >= 0.5 * region[1]:
        v[1] -= region[1]
    elif v[1] < -0.5 * region[1]:
        v[1] += region[1]        
        
    
# This function updates coordinates taking care of periodic boundaries    
def ApplyBoundaryCond():
    for n in range(nMol):
        VWrapAll(mol[n].r)

# INITIALIZE COORDINATES.
# Here a simple square lattice (with the option of unequal edge lenghts) is used,
# so that each cell contains just one atom and the system is centered about the origin
def InitCoords():

    c = np.asarray([0.0, 0.0]) # Coords
    gap = np.divide(region, initUcell)
    n = 0
    for ny in range(0, int(initUcell[1])):
        for nx in range(0, int(initUcell[0])):
            
            #c = np.asarray([nx+0.5, ny+0.5])
            #c = np.multiply(c, gap)
            #c = np.add(c, np.multiply(-0.5, region))
            #mol[n].r = c   
            
            mol[n].r = np.add(np.multiply(np.asarray([nx+0.5, ny+0.5]), gap), np.multiply(-0.5, region)) 
            n = n+1
            
            
# INITIALIZE VELOCITIES.
# The initial velocities are set to fixed magnitude (velMag)
# that depends on the temperature. After assigning random velocity directions
# the velocoties are adjusted to ensure that the center of mass is stationary.
# The function vRand serves as a source of uniformly distribuited radnom unit vectors.
def InitVels():
    
    global vSum
    vSum = np.zeros(vSum.shape)    
    
    for n in range(nMol):
        VRand(mol[n].rv)
        mol[n].rv = np.multiply(mol[n].rv, velMag)                
        vSum = np.add(vSum, mol[n].rv)


    for n in range(nMol):
        mol[n].rv = np.add(mol[n].rv, np.multiply((- 1.0 / nMol),  vSum))
        
        
# INITIALIZE ACCELERATIONS.
# The accelerations are initilized to zero
def InitAccels():
    for n in range(nMol):
        mol[n].ra = np.zeros(mol[n].ra.shape)

# Set parameters
def SetParams():

    global rCut
    global region
    global velMag # velocity magnitude
    
    rCut = math.pow(2., 1./6. * sigma)
    # Define the region
    region = np.multiply( 1./math.sqrt(density), initUcell)    
    nMol = len(mol) 
    #velocity magnitude depends on the temperature
    velMag = math.sqrt(NDIM * (1. -1. /nMol) * temperature)

        
# Setup Job
def SetupJob():
    
    global stepCount #  timestep counter 

    stepCount = 0 
    InitCoords()
    InitVels()
    InitAccels()
    AccumProps(0)

# FORCES COMPUTATION
'''
ComputeForces

ComputeForces is responsible for the interaction computations, and the interactions occur between pairs of atoms. 
The function implements the LJP, and calculates the accelerations and the forces for each pairs of atoms i and j 
located at ri and rj.
rCut = Limiting separation cutoff (rc), and it is: rCut = math.pow(2., 1./6.)
As r increases towards rCut, the force drops to 0.
Newton's third law inplies that fji = -fij, so each atom pair need only be examined once.
The amount of work is proportional to N^2.
'''

def ComputeForces():
    
    global virSum
    global uSum 
    fcVal = 0 #  The force that atom j exerts on atom i
 
    # rCut: Rc
    rrCut = Sqr(rCut)
    for n in range(nMol):
        mol[n].ra = np.zeros(mol[n].ra.shape)
    uSum = 0.
    virSum = 0.

    n = 0
    for j1 in range(nMol-1):
        for j2 in range(j1+1, nMol):
            
            # Make DeltaRij: (sum of squared RJ1-RJ2)
            dr = np.subtract(mol[j1].r, mol[j2].r) # dr contains the delta between Rj1 and Rj2
            VWrapAll(dr) # toroidal function
            rr= (dr[0] * dr[0] + dr[1] * dr[1]) # dr2
            r= np.sqrt(rr) #dr
            
            # if dr2 < Rc^2 
            if (rr < rrCut):
                rri = sigma / rr                
                rri3 = Cube(rri)
                
                # Forces calculation by Lennard-Jones potential (original from Rapaport)
                # fcVal = 48. * rri3 * (rri3 - 0.5) * rri
                # Forces calculated with the completed Lennard-Jones.
                fcVal = 48 * epsilon * np.power(sigma, 12) / np.power(r, 13) - 24 * epsilon * np.power(sigma, 6) / np.power(r, 7) 

                # Update the accelerations multiplying force for DeltaRij
                mol[j1].ra = np.add(mol[j1].ra, np.multiply(fcVal, dr))
                mol[j2].ra = np.add(mol[j2].ra, np.multiply(-fcVal, dr))
                
                # Lennard-Jones potential (original from Rapaport)
                # uSum += 4. * rri3 * (rri3 - 1.) +1. 
                # The completed Lennard-Jones.
                uSum += 4 * epsilon * np.power(sigma/r, 12)/r - np.power(sigma/r, 6) # balanced

                virSum += fcVal * rr

# INTEGRATION
'''
INTEGRATION OF COORDINATES AND VELOCITIES.
Integration of Equation of Motion uses a simple numerical techniques: the leapfrog method.
The method has excellent energy conservation properties.
LeapfrogStep integrates the coordinates and velocities. It appears twice in the listing of
SingleStep, with the argument part determinating which portion of the two-step leapfrog process
is to be performed:

vix(t + h/2) = vix(t) + (h/2)aix(t)
rix(t + h) = rix(t) + hvix (t + h/2)

'''
def LeapfrogStep(part):
    
    if part == 1:
        for n in range(nMol):
            mol[n].rv = np.add(mol[n].rv, np.multiply(0.5 * deltaT, mol[n].ra))            
            mol[n].r = np.add(mol[n].r, np.multiply(deltaT, mol[n].rv))                        
            
    else :
        for n in range(nMol):
            mol[n].rv = np.add(mol[n].rv, np.multiply(0.5 * deltaT, mol[n].ra))                        

# PROPERTIES MEASUREMENTS

def EvalProps():
    
    global vSum
    vvSum = 0.
    vSum = np.zeros(vSum.shape)
    
    global kinEnergy
    global totEnergy
    global pressure
    
    
    for n in range(nMol):
        vSum=np.add(vSum, mol[n].rv)
        vv= (mol[n].rv[0] * mol[n].rv[0] + mol[n].rv[1] * mol[n].rv[1])
        vvSum += vv
        
    kinEnergy.val = (0.5 * vvSum) / nMol
    totEnergy.val = kinEnergy.val + (uSum / nMol)
    pressure.val = density * (vvSum + virSum) / (nMol * NDIM)
    
    
    
# AccumProps functions

def PropZero(v):
    v.sum1 = v.sum2 = 0.
    return v    
    
def PropAccum(v):
    v.sum1 += v.val
    v.sum2 += Sqr(v.val)
    return v    
    
def PropAvg(v, n):
    v.sum1 /= n
    v.sum2 = math.sqrt(max(v.sum2 / n - Sqr(v.sum1), 0.)) 
    return v    
    

# AccumProps: collects results of the measurements and evaluates means and standard deviation
def AccumProps(icode):
    
    
    if icode == 0:
        PropZero(totEnergy)
        PropZero(kinEnergy)
        PropZero(pressure) 
    if icode == 1:
        PropAccum(totEnergy)
        PropAccum(kinEnergy)
        PropAccum(pressure)    
    if icode == 2:
        PropAvg(totEnergy, stepAvg)
        PropAvg(kinEnergy, stepAvg)
        PropAvg(pressure, stepAvg) 

# OUTPUT FUNCTIONS:

def plotMolCoo(mol, workdir, n):
    
    import matplotlib.patches as mpatches
    import matplotlib.pyplot as plt

    Time = timeNow
    Sigma_v = "{0:.4f}".format(vSum[0] / nMol)
    E = "{0:.4f}".format(totEnergy.sum1)
    Sigma_E = "{0:.4f}".format(totEnergy.sum2)
    Ek = "{0:.4f}".format(kinEnergy.sum1)
    Sigma_Ek = "{0:.4f}".format(kinEnergy.sum2)
    P_1 = "{0:.4f}".format(pressure.sum1)
    P_2 = "{0:.4f}".format(pressure.sum2)    
    
    TileName = (workdir+'coo/'+str(n)+'.png')

    x = []
    y = []
    
    for n in range(len(mol)):
        x.append(mol[n].r[0])
        y.append(mol[n].r[1])
        
    mark_1 = int(len(mol)/2 + len(mol)/8)
    mark_2 = int(len(mol)/2 + len(mol)/8 + 1)

    plt.cla()
    plt.plot(x, y, 'o', color='blue')
    plt.plot(x[mark_1], y[mark_1], 'o', color='red')
    plt.plot(x[mark_2], y[mark_2], 'o', color='cyan')  
    

    plt.title('timestep:'+"{0:.4f}".format(timeNow)+'; '+\
              '$\Sigma v$:'+Sigma_v+'; '+\
              'E:'+E+'; '+\
              '$\sigma E$:'+Sigma_E+';\n'+\
              'Ek:'+Ek+'; ' +\
              '$\sigma Ek$:'+Sigma_Ek+'; '+\
              'P.sum1:'+P_1+'; '+\
              'P.sum2:'+P_2+'; ', loc='left')
    
    #plt.rcParams["figure.figsize"] = (200,3)
    plt.savefig(TileName, dpi=100)
    
    
def makeMov():
    # For more information about the use of the glob package with Python, and for the convertion from 
    # gif to mp4 video formats see:    
    #https://pythonprogramming.altervista.org/png-to-gif/
    #https://stackoverflow.com/questions/6773584/how-is-pythons-glob-glob-ordered
    #https://www.programiz.com/python-programming/datetime/current-time
    #https://stackoverflow.com/questions/40726502/python-convert-gif-to-videomp4
    
    
    t = time.localtime()
    current_time = time.strftime("%D:%H:%M:%S", t)
    current_time = current_time.replace('/','-')


    # Create the frames
    frames = []
    imgs = sorted(glob.glob('coo/*.png'), key=os.path.getmtime)
    for i in imgs:
        temp = Image.open(i)
        keep = temp.copy()
        frames.append(keep)
        temp.close()

    for i in imgs:
        os.remove(i)

    # Save into a GIF file that loops forever
    frames[0].save('coo/coordinates.gif', format='GIF',
                   append_images=frames[1:],
                   save_all=True,
                   duration=30, loop=0)


    clip = mp.VideoFileClip("coo/coordinates.gif")
    clip.write_videofile("coo/"+"coordinates_"+current_time+".mp4")
    os.remove("coo/coordinates.gif")



def PrintSummary():

    print(stepCount, \
          "{0:.4f}".format(timeNow), \
          "{0:.4f}".format(vSum[0] / nMol) ,\
          "{0:.4f}".format(totEnergy.sum1),\
          "{0:.4f}".format(totEnergy.sum2), \
          "{0:.4f}".format(kinEnergy.sum1), \
          "{0:.4f}".format(kinEnergy.sum2),\
          "{0:.4f}".format(pressure.sum1),\
          "{0:.4f}".format(pressure.sum2))
    
    return (stepCount, \
          timeNow, \
          (vSum[0] / nMol) ,\
          totEnergy.sum1,\
          totEnergy.sum2, \
          kinEnergy.sum1, \
          kinEnergy.sum2,\
          pressure.sum1,\
          pressure.sum2)    


def GraphOutput():

    ax = \
    df_systemParams.plot(x="timestep", y='$\Sigma v$', kind="line")
    df_systemParams.plot(x="timestep", y='E', kind="line", ax=ax, color="C1")
    df_systemParams.plot(x="timestep", y='$\sigma E$', kind="line", ax=ax, color="C2")
    df_systemParams.plot(x="timestep",  y='Ek', kind="line", ax=ax, color="C3")
    df_systemParams.plot(x="timestep", y='$\sigma Ek$', kind="line", ax=ax, color="C4")
    df_systemParams.plot(x="timestep", y='P_1', kind="line", ax=ax, color="C9")
    df_systemParams.plot(x="timestep", y='P_2', kind="line", ax=ax, color="C9")

    plt.show()
    #plt.savefig('plot.jpg', dpi=300)

# HANDLING FUNCTION (SingleStep())
'''
SingleStep: Is the function that handles the processing for a single timestep, including: 
1) the force evaluation
2) integration of the equation of motion, 
3) adjustments required by periodic boundaries, and
4) property measurements
'''
def SingleStep():
    
    global stepCount #  timestep counter
    global timeNow    

    stepCount +=1
    timeNow = stepCount * deltaT
    LeapfrogStep(1)
    ApplyBoundaryCond()
    ComputeForces() # 1) The force evaluation
    LeapfrogStep(2) # 2) Integration of coordinates and velocities
    EvalProps()
    AccumProps(1) # Accumulate properties

    if (stepCount % stepAvg == 0):
        AccumProps(2) # Calculate averages
        systemParams.append(PrintSummary())
        AccumProps(0) # Set to zero all the properties.

# 2D SOFT-DISK SIMULATION: THE MAIN LOOP

# Import libraries for system operations
import os.path
from os import path
import shutil

class Mol():
    def __init__(self, r, rv, ra):
        self.r = np.asarray([0.0, 0.0]) 
        self.rv = np.asarray([0.0, 0.0])
        self.ra = np.asarray([0.0, 0.0])
        
        
class Prop():
    def __init__(self, val, sum1, sum2 ):
        self.val=val
        self.sum1=sum1
        self.sum2=sum2       

# PARAMETERS
mov = 1 # set mov=1 if you want make a video

# Set a working directory for all the png and videos
workdir = str(os.getcwd()+'/')

# If the /coo directory doesn't exist make it, else remove /coo (and its contents) and 
# create a new /coo directory.
if path.exists(str(workdir+'coo'))==False:
    os.makedirs(str(workdir+'coo'))
else:
    shutil.rmtree(str(workdir+'coo'))
    os.makedirs(str(workdir+'coo'))

# Load the input parameter file
df_params = pd.read_csv('Rap_2_LJP.in', sep='\t', header=None, names=['parameter', 'value'])

NDIM = 2 # Two-Dimension setting
vSum = np.asarray([0.0, 0.0]) # velocity sum
kinEnergy =Prop(0.0, 0.0, 0.0) #Ek (and average)
totEnergy =Prop(0.0, 0.0, 0.0) #E (and average)
pressure  =Prop(0.0, 0.0, 0.0) #P (and average) 

systemParams = []

IADD = 453806245
IMUL = 314159269
MASK = 2147483647
SCALE = 0.4656612873e-9
randSeedP = 17

deltaT = float(df_params.values[0][1])
density = float(df_params.values[1][1])

initUcell = np.asarray([0.0, 0.0]) # initialize cell
initUcell[0] = int(df_params.values[2][1])
initUcell[1] = int(df_params.values[3][1])

stepAvg = int(df_params.values[4][1])
stepEquil = float(df_params.values[5][1])
stepLimit = float(df_params.values[6][1])
temperature = float(df_params.values[7][1])
float(df_params.values[7][1])

#Define an array of Mol
mol = [Mol(np.asarray([0.0, 0.0]), \
           np.asarray([0.0, 0.0]), \
           np.asarray([0.0, 0.0])) for i in range(int(initUcell[0]*initUcell[1]))]


# Define the number of molecules
global nMol
nMol = len(mol)

# LJP parameters:
epsilon =  1
sigma = 1


# START THE MAIN LOOP
SetParams()
SetupJob()
moreCycles = 1

n = 0
while moreCycles:
    SingleStep()
    if mov==1:
        plotMolCoo(mol, workdir, n) # Make a graph of the coordinates
    n += 1
    if stepCount >= stepLimit:
        moreCycles = 0
        

columns = ['timestep','timeNow', '$\Sigma v$', 'E', '$\sigma E$', 'Ek', '$\sigma Ek$', 'P_1', 'P_2']
df_systemParams = pd.DataFrame(systemParams, columns=columns)        

# Make a video
if mov==1:
    makeMov()

GraphOutput()





# -*- coding: utf-8 -*-
"""

    ファイル名：   GuiSettingInfo.py

    処理内容：

        DISyn GUI のパラメータ入力フレーム

        Copyright (c) 2016-2021, SAXS Team, KEK-PF

"""
from __future__ import division, print_function, unicode_literals

import sys
import os
import re

from molass_legacy.KekLib.OurTkinter         import Tk, FileDialog, ToolTip
import OurMessageBox        as MessageBox

import PilatusCounter
from PilatusUtils       import get_in_folder_info
import ChangeableLogger
from CreateFolderDialog import CreateFolderDialog
from molass_legacy.KekLib.BasicUtils         import exe_name
from Preferences        import get_preference
from Development        import get_devel_info
from SynthesizerSettings    import clear_settings, get_setting, set_setting, set_mask, get_mask
from molass_legacy.KekLib.TkSupplements      import SlimButton

DEFAULT_WATCH_INVERVAL = 180

def is_empty_val( val ):
    return val == None or val == '' or val[0] == '<' or val == 'None' 

class EntryFrame(Tk.Frame):

    def __init__(self, master, controller):
        Tk.Frame.__init__(self, master)

        self.during_init = True
        self.controller = controller
        self.entries    = []
        self.app_logger = None
        self.pilatus_counter    = None
        self.auto_set_asked = False
        self.auto_set_done  = False
        self.adj_output     = get_devel_info( 'adj_output' )
        self.data_entry_ready = False
        self.in_construction = True
        self.logfile_path   = ''
        self.error_notified = {}

        grid_row = -1

        grid_row += 1
        in_folder_name  = Tk.Label( self, text= 'Measured Image Folder: ' )
        in_folder_name.grid( row=grid_row, column=0, sticky=Tk.E )
        self.in_folder = Tk.StringVar()
        in_folder = get_setting( 'in_folder' )
        self.in_folder.set( in_folder )
        self.in_folder_entry = Tk.Entry( self, textvariable=self.in_folder, width=80 )
        self.entries.append( [ 'dir', self.in_folder_entry, self.on_entry_in_folder ] )
        self.in_folder_entry.grid( row=grid_row, column=1 )
        b1 = Tk.Button( self, text='...', command=self.select_in_folder )
        b1.grid( row=grid_row, column=2, sticky=Tk.W )
        ToolTip( b1, 'Select an appropriate folder with this button.' )
        cm1 = Tk.Label( self, text= ' must be entered manually' )
        cm1.grid( row=grid_row, column=3, sticky=Tk.W )

        grid_row += 1
        log_file_name = Tk.Label( self, text= 'Measurement Log File: ' )
        log_file_name.grid( row=grid_row, column=0, sticky=Tk.E )
        self.log_file = Tk.StringVar()
        log_file = get_setting( 'log_file' )
        self.log_file.set( log_file )
        self.log_file_entry = Tk.Entry( self, textvariable=self.log_file, width=80 )
        self.entries.append( [ 'file', self.log_file_entry, self.on_entry_log_file ] )
        self.log_file_entry.grid( row=grid_row, column=1 )
        b2 = Tk.Button( self, text='...', command=self.select_log_file )
        b2.grid( row=grid_row, column=2, sticky=Tk.W )
        ToolTip( b2, 'Select an appropriate file with this button.' )
        cm2 = Tk.Label( self, text= ' automatically set if exists in the above folder' )
        cm2.grid( row=grid_row, column=3, sticky=Tk.W )

        grid_row += 1
        mask_file_name = Tk.Label( self, text= 'SAngler Mask File: ' )
        mask_file_name.grid( row=grid_row, column=0, sticky=Tk.E  )
        self.mask_file = Tk.StringVar()
        mask_file = get_setting( 'mask_file' )
        self.mask_file.set( mask_file )
        self.mask_file_entry = Tk.Entry( self, textvariable=self.mask_file, width=80 )
        self.entries.append( [ 'file', self.mask_file_entry, self.on_entry_mask_file ] )
        self.mask_file_entry.grid( row=grid_row, column=1 )
        b3 = Tk.Button( self, text='...', command=self.select_mask_file )
        b3.grid( row=grid_row, column=2, sticky=Tk.W )
        ToolTip( b3, 'Select an appropriate file with this button.' )
        cm3 = Tk.Label( self, text= ' automatically set if exists in the above folder' )
        cm3.grid( row=grid_row, column=3, sticky=Tk.W )

        if self.adj_output == 'YES':
            grid_row += 1
            adj_folder_name = Tk.Label( self, text= 'Adjusted Image Folder: ' )
            adj_folder_name.grid( row=grid_row, column=0, sticky=Tk.E  )
            self.adj_folder = Tk.StringVar()
            adj_folder = get_setting( 'adj_folder' )
            self.adj_folder.set( adj_folder )
            self.adj_folder_entry = Tk.Entry( self, textvariable=self.adj_folder, width=80 )
            self.entries.append( [ 'dir', self.adj_folder_entry, self.on_entry_adj_folder ] )
            self.adj_folder_entry.grid( row=grid_row, column=1 )
            b4 = Tk.Button( self, text='...', command=self.select_adj_folder )
            b4.grid( row=grid_row, column=2, sticky=Tk.W )
            ToolTip( b4, 'Select an appropriate folder with this button.' )
            cm4 = Tk.Label( self, text= ' must be entered manually' )
            cm4.grid( row=grid_row, column=3, sticky=Tk.W )

        grid_row += 1
        syn_folder_name = Tk.Label( self, text= 'Synthesized Image Folder: ' )
        syn_folder_name.grid( row=grid_row, column=0, sticky=Tk.E  )
        self.syn_folder = Tk.StringVar()
        syn_folder = get_setting( 'syn_folder' )
        self.syn_folder.set( syn_folder )
        self.syn_folder_entry = Tk.Entry( self, textvariable=self.syn_folder, width=80 )
        self.entries.append( [ 'dir', self.syn_folder_entry, self.on_entry_syn_folder ] )
        self.syn_folder_entry.grid( row=grid_row, column=1 )
        b5 = Tk.Button( self, text='...', command=self.select_syn_folder )
        b5.grid( row=grid_row, column=2, sticky=Tk.W )
        ToolTip( b5, 'Select an appropriate folder with this button.' )
        cm5 = Tk.Label( self, text= ' must be entered manually' )
        cm5.grid( row=grid_row, column=3, sticky=Tk.W )

        grid_row += 1
        op_option_label = Tk.Label( self, text= 'Operation Mode: ' )
        op_option_label.grid( row=grid_row, column=0, sticky=Tk.E  )
        op_option_frame = Tk.Frame( self )
        op_option_frame.grid( row=grid_row, column=1, sticky=Tk.W  )
        self.op_option = Tk.StringVar()
        j = 0
        spaces = '    '
        self.op_option_buttons = []
        for t, v in ( [ 'Manual   ' + spaces, 'MANUAL' ], [ 'Automatic' + spaces, 'AUTOMATIC' ] ):
            b = Tk.Radiobutton( op_option_frame, text=t,
                            variable=self.op_option, value=v )
            self.op_option_buttons.append( b )
            b.grid( row=0, column=j, sticky=Tk.W )
            j += 1

        self.watch_interval_label = Tk.Label( op_option_frame, text= 'Watch Interval ' )
        self.watch_interval_label.grid( row=0, column=j, sticky=Tk.W )

        j += 1
        self.watch_interval = Tk.StringVar()
        watch_interval = get_setting( 'watch_interval' )
        self.watch_interval.set( str( watch_interval ) )
        self.watch_interval_box = Tk.Spinbox( op_option_frame, textvariable=self.watch_interval,
                                                from_=10, to=600, increment=10, width=4, justify=Tk.CENTER )
        self.watch_interval_box.grid( row=0, column=j, sticky=Tk.W )

        j += 1
        self.watch_interval_s = Tk.Label( op_option_frame, text= 's  ' )
        self.watch_interval_s.grid( row=0, column=j, sticky=Tk.W )

        j += 1
        self.autorun_button = SlimButton( op_option_frame, text='Auto-Run Start', command=self.controller.auto_start )
        self.autorun_button.grid( row=0, column=j, sticky=Tk.W )
        ToolTip( self.autorun_button, "Select 'Automatic' and press this button to start automatic operation." )

        self.previous_syn_folder    = None
        if not is_empty_val(syn_folder):
            self.on_entry_syn_folder()

        self.add_focus_event_binds()

        self.add_dnd_bind()
        self.during_init = False

    def after_construction_proc( self ):
        """
            このメソッドは常に __init__ に続けて実行するものであるが、
            テストの場合、mainloop に入った後に実行する必要が生じることがあるので
            分離している。
        """
        # 入力項目の初期値をチェックし、従属する処理があれば実行する。
        check_result = self.check_entries()
        print( 'check_result=', check_result )

        self.previous_in_folder     = None
        self.previous_log_file      = None
        self.previous_mask_file     = None
        self.previous_adj_folder    = None

        if check_result:
            # TODO: refactoring on_entry's
            self.on_entry_in_folder( event=None, check_only=True )
            self.on_entry_log_file()
            self.on_entry_mask_file()
            if self.adj_output == 'YES':
                self.on_entry_adj_folder()
            self.on_entry_syn_folder()
            # これらの入力値は妥当であり、
            # 初期には、それによってテーブルが初期化されるため、
            # Refresh の注意喚起は不要
            self.controller.refresh_button_suggestion.stop()

        # must be after on_entry's
        self.previous_in_folder     = self.in_folder.get()
        self.previous_log_file      = self.log_file.get()
        self.previous_mask_file     = self.mask_file.get()
        if self.adj_output == 'YES':
            self.previous_adj_folder    = self.adj_folder.get()
        self.previous_syn_folder    = self.syn_folder.get()

        # 初期値設定による変更は変更としない。
        self.changed_after_refresh  = False
        self.changed_ever           = False
        # print '__init__: changed_after_refresh=', self.changed_after_refresh

        # 以降、値の変更を監視する。
        def entry_tracer( *args ):
            # print 'entry_tracer:', 
            self.changed_after_refresh  = True
            self.changed_ever           = True

        self.in_folder.trace ( 'w', entry_tracer )
        self.log_file.trace  ( 'w', entry_tracer )
        self.mask_file.trace ( 'w', entry_tracer )
        if self.adj_output == 'YES':
            self.adj_folder.trace( 'w', entry_tracer )
        self.syn_folder.trace( 'w', entry_tracer )

        def op_option_tracer( *args ):
            op_option_ = self.op_option.get()
            set_setting( 'op_option', op_option_ )
            if op_option_ == 'MANUAL':
                self.auto_run_widgets_disable()
                self.controller.auto_disable( force=True )
            else:
                
                if self.check_entries():
                    self.auto_run_widgets_enable()
                    self.controller.auto_enable()
                else:
                    if not self.in_construction:
                        MessageBox.showinfo( "Not Allowed", "You can't select 'Automatic' until all required entries are filled." )
                        self.op_option.set( 'MANUAL' )
                        self.op_option_buttons[1].event_generate( "<Leave>" )
                        self.op_option_buttons[0].event_generate( "<Button-1>" )
                        self.update()
                        # self.error_entries[0].focus_force()

        self.op_option.trace( 'w', op_option_tracer )

        self.op_option.set( get_setting( 'op_option' ) )

        self.in_construction = False

    def __del__( self ):
        # print( 'EntryFrame.__del__' )
        self.app_logger = None

    def clear( self ):
        self.clear_entries()
        self.check_entries()        # neccesary as of 2015-12-09

        self.previous_in_folder     = ''
        self.previous_log_file      = ''
        self.previous_mask_file     = ''
        if self.adj_output == 'YES':
            self.previous_adj_folder    = ''
        self.previous_syn_folder    = ''
        set_setting( 'in_folder', '' )
        set_setting( 'log_file', '' )
        set_setting( 'mask_file', '' )
        if self.adj_output == 'YES':
            set_setting( 'adj_folder', '' )
        set_setting( 'syn_folder', '' )

    def auto_run_widgets_disable( self ):
        self.watch_interval_label.configure( state='disabled' )
        self.watch_interval_box.configure( state='disabled' )
        self.watch_interval_s.configure( state='disabled' )
        self.autorun_button.configure( state='disabled' )

    def auto_run_widgets_enable( self ):
        self.watch_interval_label.configure( state='normal' )
        self.watch_interval_box.configure( state='normal' )
        self.watch_interval_s.configure( state='normal' )
        self.autorun_button.configure( state='normal' )

    def add_focus_event_binds( self ):
        def suggest_text_and_check( event, text='', proc=None ):
            current = event.widget.get()
            # print 'suggest_text: current=', current, '; entry=', on_entry
            if current == '':
                event.widget.insert( 0, text )
                event.widget.configure( fg='red' )
            elif current[0] == '<':
                event.widget.delete( 0, Tk.END )

            # この proc には、self.entries に登録された on_entry が
            # 渡されている。
            if proc:
                proc()

        for obj_type, entry, on_entry in self.entries:
            current = entry.get()
            # print 'add_suggest_text_bind: current=', current

            if is_empty_val( current ):
                entry.delete( 0, Tk.END )
                if obj_type == 'dir':
                    entry.insert( 0, '<Folder>' )
                else:
                    entry.insert( 0, '<File>' )
            else:
                pass

            entry.bind( "<FocusIn>",    lambda event: suggest_text_and_check( event ) )

            # 以下の lambda では、on_entry を値渡しするために default argument を使用している。
            # 参照URL: http://stackoverflow.com/questions/10452770/python-lambdas-binding-to-local-values
            # また、lambda 表現の制限により、suggest_text_and_check に２つの処理を実装している。
            if obj_type == 'dir':
                entry.bind( "<FocusOut>",   lambda event, proc_value=on_entry:
                                                suggest_text_and_check( event, '<Folder>', proc_value ) )
            else:
                entry.bind( "<FocusOut>",   lambda event, proc_value=on_entry:
                                                suggest_text_and_check( event, '<File>', proc_value ) )

    def check_entries( self, index=-999 ):
        """
            Drag and Drop の場合はフォルダやファイルを指定できないので、
            （存在はしても）違ったタイプのパスが設定されることがある点に
            注意。
        """
        # print( 'check_entries: index=', index )
        msg = ''
        ret = True
        if index == -999:
            self.error_entries = []

        if index >= -1:
            entries_ = [ self.entries[index] ]
        else:
            entries_ = self.entries

        error_index = None
        i_ = -1
        for obj_type, entry, on_entry in entries_:
            i_ += 1
            path = entry.get()
            entry_ok = True
            # print 'Checking if %s is a %s' % ( path, obj_type )
            if path and os.path.exists( path ):
                if obj_type == 'dir':
                    if os.path.isdir( path ):
                        entry.configure( fg='black' )
                    else:
                        if msg == '':
                            error_index = i_
                            msg = '%s is not a folde.r' % ( path )
                        entry.configure( fg='red' )
                        ret = False
                        entry_ok = False
                elif obj_type == 'file':
                    if os.path.isfile( path ):
                        entry.configure( fg='black' )
                    else:
                        if msg == '':
                            error_index = i_
                            msg = '%s is not a file.' % ( path )
                        entry.configure( fg='red' )
                        ret = False
                        entry_ok = False
                else:
                    pass

            else:
                if not is_empty_val( path ):
                    if msg == '':
                        error_index = i_
                        msg = '%s does not exist.' % ( path )

                entry.configure( fg='red' )
                ret = False
                entry_ok = False

            if index == -999:
                if not entry_ok:
                    self.error_entries.append( entry )

        if msg:
            if index == -999:
                if error_index == len( entries_ ) - 1:
                    # Synthesized Image Folder is referred to as -1
                    error_index = -1
                self.error_notified[ ( error_index, msg ) ] = True
                reply = MessageBox.askyesno( 'Entry Error & Clear Info Question',
                                msg
                                + "\nLooks like setting environment has been changed."
                                + "\nDo you want to clear the previous setting info?"
                                )
                # print( reply )
                if reply:
                    clear_settings()
                    self.clear_entries()
                    self.check_entries()
                else:
                    pass
            else:
                if self.error_notified.get( ( index, msg ) ) is None:
                    MessageBox.showerror( 'Entry Error', msg )
                self.error_notified[ ( index, msg ) ] = True

        # print "check_entries: ret=", ret
        if ret:
            if index == -999:
                if get_mask() == None:
                    # mask が設定されていない場合がありうるので、
                    # 強制的に設定する。
                    print( 'mask is not set yet' )
                    self.on_entry_mask_file( set_force=True )
                self.data_entry_ready = True
                if self.op_option.get() == 'MANUAL':
                    if self.controller.op_is_manual:
                        print( '(1) calling controller.auto_disable()' )
                        self.controller.auto_disable()
                else:
                    if not self.controller.op_is_manual:
                        print( '(2) calling controller.auto_enable()' )
                        self.controller.auto_enable()
            else:
                pass

            # if get_setting( 'op_option' ) == 'MANUAL':
            # TODO: set_setting( 'op_option' )
            if self.op_option.get() == 'MANUAL':
                all_entries_are_filled = True
                for obj_type, entry, on_entry in self.entries:
                    if is_empty_val( entry.get() ):
                        all_entries_are_filled = False
                        break
                if all_entries_are_filled:
                    self.controller.refresh_button_enable()
                    if self.changed_after_refresh:
                        self.controller.refresh_button_suggestion.start()
        else:
            self.data_entry_ready = False
            if self.controller.op_is_manual:
                print( '(3) calling controller.auto_disable()' )
                if self.during_init:
                    pass
                else:
                    self.controller.auto_disable()
                    self.controller.manual_buttons_disable()
                    self.controller.refresh_button_suggestion.stop()

        print( 'check_entries: index=%d, ret=%s' % ( index, ret ) )

        return ret

    def clear_entries( self ):
        for obj_type, entry, on_entry in self.entries:
            entry.delete( 0, Tk.END )
            # TODO: consistency in the empty prompts
            if obj_type == 'dir':
                entry.insert( 0, '<Folder>' )
            else:
                entry.insert( 0, '<File>' )

    def add_dnd_bind( self ):
        self.in_folder_entry.register_drop_target("*")
        self.log_file_entry.register_drop_target("*")
        self.mask_file_entry.register_drop_target("*")
        self.syn_folder_entry.register_drop_target("*")

        def dnd_handler( event ):
            event.widget.delete( 0, Tk.END )
            data = re.sub( r'({|})', '', event.data )
            event.widget.insert( 0, data )

        def dnd_handler_in_folder( event ):
            dnd_handler( event )
            self.on_entry_in_folder()

        def dnd_handler_log_file( event ):
            dnd_handler( event )
            self.on_entry_log_file()

        def dnd_handler_mask_file( event ):
            dnd_handler( event )
            self.on_entry_mask_file()

        def dnd_handler_adj_folder( event ):
            dnd_handler( event )
            self.on_entry_adj_folder()

        def dnd_handler_syn_folder( event ):
            dnd_handler( event )
            self.on_entry_syn_folder()

        self.in_folder_entry.bind("<<Drop>>", dnd_handler_in_folder)
        self.log_file_entry.bind("<<Drop>>", dnd_handler_log_file)
        self.mask_file_entry.bind("<<Drop>>", dnd_handler_mask_file)

        if self.adj_output == 'YES':
            self.adj_folder_entry.register_drop_target("*")
            self.adj_folder_entry.bind("<<Drop>>", dnd_handler_adj_folder)
        self.syn_folder_entry.bind("<<Drop>>", dnd_handler_syn_folder)

        for obj_type, entry, on_entry in self.entries:
            ToolTip( entry, 'You can enter directly, or use the right file dialog button, otherwise, drag and drop here.')

    def askdirectory( self, entry_variable, suggest_name=False ):
        entered_path = entry_variable.get()
        # print 'entered_path=', entered_path
        dir_ = os.path.dirname( entered_path ).replace( '/', '\\' )
        if suggest_name:
            if is_empty_val( entered_path ):
                dialog = CreateFolderDialog( self, 'Folder Creation Dialog', self.suggested_folders( suggest_name ) )
                if dialog.created_folder:
                    return dialog.created_folder

        # print 'dir_=', dir_
        f = FileDialog.askdirectory( initialdir=dir_ )
        return f

    def suggested_folders( self, name ):
        folders = []
        in_folder = self.in_folder.get()
        if not is_empty_val( in_folder ):
            in_path = os.path.abspath( in_folder )
            f = os.path.join( in_path, name )
            folders.append( f )
            f = in_path + '-' + name
            folders.append( f )

        cwd_ = os.getcwd()

        drive_ = cwd_.split( '\\' )[0]
        f = os.path.join( drive_ + '\\', name )
        folders.append( f )

        ret_foleds = []
        for f in folders:
            ret_foleds.append( f.replace( '\\', '/' ) )
        return ret_foleds

    def select_in_folder( self ):
        f = self.askdirectory( self.in_folder )
        if not f:
            return

        self.in_folder.set( f )
        print( 'select_in_folder: self.on_entry_in_folder call' )
        self.on_entry_in_folder()

    def on_entry_in_folder( self, event=None, check_only=False ):
        # print( 'on_entry_in_folder: check_only=', check_only )
        self.auto_set_asked = False
        self.auto_set_done  = False

        in_folder = self.in_folder.get()

        if not self.check_entries( 0 ):
            self.previous_in_folder       = in_folder
            return

        if check_only:
            return

        if self.previous_in_folder and in_folder == self.previous_in_folder:
            # 値が同じならば以下の処理は不要
            return
        self.previous_in_folder       = in_folder
        set_setting( 'in_folder', in_folder )

        in_folder = self.in_folder.get()
        log_file, mask_file = get_in_folder_info( in_folder )

        do_auto_fill = True
        if ( log_file or mask_file ):
            if is_empty_val( self.log_file.get() ):
                action_ = 'Insert'
            else:
                action_ = 'Replace'
            self.auto_set_asked = True
            yes = MessageBox.askyesno( 'Quention', 'Log and/or mask files exist in the folder. %s the corresponding entries?' % (action_) )
            if not yes:
                do_auto_fill =False

        if do_auto_fill:
            if log_file:
                log_file_   = '%s/%s' % ( in_folder, log_file )
                self.log_file.set( log_file_ )
                self.auto_set_done  = True
                self.on_entry_log_file()
            if mask_file:
                mask_file_  = '%s/%s' % ( in_folder, mask_file)
                self.auto_set_done  = True
                self.mask_file.set( mask_file_ )
                self.on_entry_mask_file()

        if self.controller.image_info_table is not None:
            # print( 'on_entry_in_folder: current_data_end=', self.controller.image_info_table.current_data_end )
            # in_folder が更新されたので、カレントをリセットする。
            self.controller.image_info_table.current_data_end = 0

    def select_log_file(self):
        f = FileDialog.askopenfilename( initialdir=get_setting( 'in_folder' ) )
        if not f:
            return

        self.log_file.set( f )
        self.on_entry_log_file()

    def on_entry_log_file( self ):
        f = self.log_file.get()
        if not self.check_entries( 1 ):
            self.previous_log_file = f
            return

        if self.previous_log_file and f == self.previous_log_file:
            # 値が同じならば以下の処理は不要
            return

        self.previous_log_file = f
        self.pilatus_counter = PilatusCounter.Counter( self.in_folder.get() )
        set_setting( 'log_file', f )

    def select_mask_file(self):
        f = FileDialog.askopenfilename( initialdir=get_setting( 'in_folder' ) )
        if not f:
            return

        self.mask_file.set( f )
        self.on_entry_mask_file()

    def on_entry_mask_file( self, set_force=False ):
        f = self.mask_file.get()

        if not self.check_entries( 2 ):
            self.previous_mask_file = f
            return

        if not set_force and self.previous_mask_file and f == self.previous_mask_file:
            # 値が同じならば以下の処理は不要
            return

        self.previous_mask_file = f
        try:
            print( 'on_entry_mask_file: set_mask, set_force=', set_force )
            if set_mask( f ):
                self.mask_file_entry.configure( fg='black' )
                print( get_mask() )
            else:
                # TODO: get the error msg from SAnglerMask
                self.mask_file_entry.configure( fg='red' )
                MessageBox.showerror( 'SAngler Mask File Error', 'Not a well formatted SAngler mask file' )
                return False

            set_setting( 'mask_file', f )
        except Exception as err:
            print( 'on_entry_mask_file: unexpected exception', err.message )
            return False

        return True

    def select_adj_folder(self):
        f = self.askdirectory( self.adj_folder, suggest_name='Adjusted' )
        if not f:
            return
        self.adj_folder.set( f )
        self.on_entry_adj_folder()

    def on_entry_adj_folder( self ):
        f = self.adj_folder.get()
        if not self.check_entries( 3 ):
            self.previous_adj_folder    = f
            return

        if self.previous_adj_folder and f == self.previous_adj_folder:
            # 値が同じならば以下の処理は不要
            return

        self.previous_adj_folder = f
        set_setting( 'adj_folder', f )

    def select_syn_folder(self):
        f = self.askdirectory( self.syn_folder, suggest_name='Synthesized' )
        if not f:
            return

        self.syn_folder.set( f )
        self.on_entry_syn_folder()

    def on_entry_syn_folder( self ):
        f = self.syn_folder.get()
        if not self.check_entries( -1 ):
            self.previous_syn_folder = f
            return

        if self.previous_syn_folder and f == self.previous_syn_folder:
            # 値が同じならば以下の処理は不要
            return

        self.previous_syn_folder = f
        self.on_entry_syn_folder_sub()

    def on_entry_syn_folder_sub( self ):
        syn_folder = self.syn_folder.get()
        if is_empty_val( syn_folder ):
            return True

        set_setting( 'syn_folder', syn_folder )
        exe_name_ =  exe_name()
        self.logfile_path = '%s/%s.log' % ( syn_folder, exe_name_ )
        if self.app_logger:
            self.app_logger.changeto( self.logfile_path )
        else:
            self.app_logger = None
            self.app_logger = ChangeableLogger.Logger( self.logfile_path )

        print( 'on_entry_syn_folder_sub: return' )
        return True

    def change_reset( self ):
        self.changed_after_refresh   = False

    def has_been_changed_after_refresh( self ):
        # print 'has_been_changed_after_refresh: changed_after_refresh=', self.changed_after_refresh
        return self.changed_after_refresh

    def has_been_changed_ever( self ):
        return self.changed_ever

    """
        以下は、AutoRunController から起動した ActionWindow が、
        終了時（OK を押したとき）に grab が効かない状態になるための
        暫定措置として使用する。（ grab が有効ならば不要 ）
    """
    def disable( self ):
        for obj_type, entry, on_entry in self.entries:
            entry.configure( state='disabled' )

        for button in self.op_option_buttons:
            button.configure( state='disabled' )

        if self.op_option.get() == 'MANUAL':
            # manual の場合 grab が有効のため、ここは未使用。
            self.controller.refresh_button_disable()
            self.controller.run_button_disable()
        else:
            self.auto_run_widgets_disable()

    def enable( self ):
        print( 'EntryFrame: enable' )
        for obj_type, entry, on_entry in self.entries:
            entry.configure( state='normal' )

        for button in self.op_option_buttons:
            button.configure( state='normal' )

        if self.op_option.get() == 'MANUAL':
            # manual の場合 grab が有効のため、ここは未使用。
            self.controller.refresh_button_enable()
            self.controller.run_button_enable()
        else:
            self.auto_run_widgets_enable()
